"""Module for modeling run parameters file parsing."""
import logging
from pathlib import Path
from typing import Optional, Dict
from xml.etree import ElementTree

from cg.constants.demultiplexing import RunParametersXMLNodes
from cg.constants.sequencing import Sequencers
from cg.io.xml import read_xml
from cg.exc import RunParametersError

LOG = logging.getLogger(__name__)


class RunParameters:
    """Base class with basic functions to handle the run parameters from a sequencing run."""

    def __init__(self, run_parameters_path: Path):
        self.path: Path = run_parameters_path
        self.tree: ElementTree = read_xml(file_path=run_parameters_path)
        self.validate_instrument()

    @property
    def index_length(self) -> int:
        """Return the length of the indexes if they are equal, raise an error otherwise."""
        index_one_length: int = self.get_index_1_cycles()
        index_two_length: int = self.get_index_2_cycles()
        if index_one_length != index_two_length:
            raise RunParametersError("Index lengths are not the same!")
        return index_one_length

    @staticmethod
    def node_not_found(node: Optional[ElementTree.Element], name: str) -> None:
        """Raise exception if the given node is not found."""
        if node is None:
            message = f"Could not determine {name}"
            LOG.warning(message)
            raise RunParametersError(message)

    def validate_instrument(self) -> None:
        """Raise an error if the parent class was instantiated."""
        raise NotImplementedError(
            "Parent class instantiated. Instantiate instead RunParametersNovaSeq6000 or RunParametersNovaSeqX"
        )

    @property
    def requires_dummy_samples(self) -> Optional[bool]:
        """Return true if the flow cell requires the addition of dummy samples."""
        raise NotImplementedError("Impossible to know dummy sample requirements of parent class")

    @property
    def control_software_version(self) -> Optional[str]:
        """Return the control software version if existent."""
        raise NotImplementedError(
            "Impossible to retrieve control software version from parent class"
        )

    @property
    def reagent_kit_version(self) -> Optional[str]:
        """Return the reagent kit version if existent."""
        raise NotImplementedError("Impossible to retrieve reagent kit version from parent class")

    @property
    def sequencer(self) -> Optional[str]:
        """Return the sequencer associated with the current run parameters."""
        raise NotImplementedError("Impossible to retrieve sequencer from parent class")

    def get_index_1_cycles(self) -> int:
        """Return the number of cycles in the first index read."""
        raise NotImplementedError("Impossible to retrieve index1 cycles from parent class")

    def get_index_2_cycles(self) -> int:
        """Return the number of cycles in the second index read."""
        raise NotImplementedError("Impossible to retrieve index2 cycles from parent class")

    def get_read_1_cycles(self) -> int:
        """Return the number of cycles in the first read."""
        raise NotImplementedError("Impossible to retrieve read1 cycles from parent class")

    def get_read_2_cycles(self) -> int:
        """Return the number of cycles in the second read."""
        raise NotImplementedError("Impossible to retrieve read2 cycles from parent class")

    def __str__(self):
        return f"RunParameters(path={self.path}, sequencer={self.sequencer})"

    def __repr__(self):
        return (
            f"RunParameters(path={self.path},"
            f"reagent_kit_version={self.reagent_kit_version},"
            "control_software_version={self.control_software_version},"
            f"index_length={self.index_length},"
            f"sequencer={self.sequencer})"
        )


class RunParametersNovaSeq6000(RunParameters):
    """Specific class for parsing run parameters of NovaSeq6000 sequencing."""

    def validate_instrument(self) -> None:
        """Raise an error if the class was not instantiated with a NovaSeq6000 file."""
        node_name: str = RunParametersXMLNodes.APPLICATION
        xml_node: Optional[ElementTree.Element] = self.tree.find(node_name)
        self.node_not_found(node=xml_node, name="Instrument")
        if xml_node.text != RunParametersXMLNodes.NOVASEQ_6000_APPLICATION:
            raise RunParametersError(
                "The file parsed does not correspond to a NovaSeq6000 instrument"
            )

    @property
    def requires_dummy_samples(self) -> bool:
        """Return true if the number of cycles of both indexes is 8."""
        return self.index_length != 8

    @property
    def control_software_version(self) -> str:
        """Return the control software version."""
        node_name: str = RunParametersXMLNodes.APPLICATION_VERSION
        xml_node: Optional[ElementTree.Element] = self.tree.find(node_name)
        self.node_not_found(node=xml_node, name="control software version")
        return xml_node.text

    @property
    def reagent_kit_version(self) -> str:
        """Return the reagent kit version if existent, return 'unknown' otherwise."""
        node_name: str = RunParametersXMLNodes.REAGENT_KIT_VERSION
        xml_node: Optional[ElementTree.Element] = self.tree.find(node_name)
        if xml_node is None:
            LOG.warning("Could not determine reagent kit version")
            LOG.info("Set reagent kit version to 'unknown'")
            return RunParametersXMLNodes.UNKNOWN_REAGENT_KIT_VERSION
        return xml_node.text

    @property
    def sequencer(self) -> str:
        """Return the sequencer associated with the current run parameters."""
        return Sequencers.NOVASEQ.value

    def get_node_integer_value(self, node_name: str, name: str) -> int:
        """Return the value of the node as an integer."""
        xml_node = self.tree.find(node_name)
        self.node_not_found(node=xml_node, name=name)
        return int(xml_node.text)

    def get_index_1_cycles(self) -> int:
        """Return the number of cycles in the first index read."""
        node_name = RunParametersXMLNodes.INDEX_1_NOVASEQ_6000
        return self.get_node_integer_value(node_name=node_name, name="length of index one")

    def get_index_2_cycles(self) -> int:
        """Return the number of cycles in the second index read."""
        node_name = RunParametersXMLNodes.INDEX_2_NOVASEQ_6000
        return self.get_node_integer_value(node_name=node_name, name="length of index two")

    def get_read_1_cycles(self) -> int:
        """Return the number of cycles in the first read."""
        node_name = RunParametersXMLNodes.READ_1_NOVASEQ_6000
        return self.get_node_integer_value(node_name=node_name, name="length of reads one")

    def get_read_2_cycles(self) -> int:
        """Return the number of cycles in the second read."""
        node_name = RunParametersXMLNodes.READ_2_NOVASEQ_6000
        return self.get_node_integer_value(node_name=node_name, name="length of reads two")


class RunParametersNovaSeqX(RunParameters):
    """Specific class for parsing run parameters of NovaSeqX sequencing."""

    def validate_instrument(self) -> None:
        """Raise an error if the class was not instantiated with a NovaSeqX file."""
        node_name: str = RunParametersXMLNodes.INSTRUMENT_TYPE
        xml_node: Optional[ElementTree.Element] = self.tree.find(node_name)
        self.node_not_found(node=xml_node, name="Instrument")
        if xml_node.text != RunParametersXMLNodes.NOVASEQ_X_INSTRUMENT:
            raise RunParametersError("The file parsed does not correspond to a NovaSeqX instrument")

    @property
    def requires_dummy_samples(self) -> bool:
        """Return False for run parameters associated with NovaSeqX sequencing."""
        return False

    @property
    def control_software_version(self) -> None:
        """Return None for run parameters associated with NovaSeqX sequencing."""
        return

    @property
    def reagent_kit_version(self) -> None:
        """Return None for run parameters associated with NovaSeqX sequencing."""
        return

    @property
    def sequencer(self) -> str:
        """Return the sequencer associated with the current run parameters."""
        return Sequencers.NOVASEQX.value

    @property
    def read_parser(self) -> Dict[str, int]:
        """Return read and index cycle values parsed as a dictionary."""
        cycle_mapping: Dict[str, int] = {}
        planned_reads: Optional[ElementTree.Element] = self.tree.find(
            RunParametersXMLNodes.PLANNED_READS
        )
        self.node_not_found(node=planned_reads, name="PlannedReads")
        read_elem: ElementTree.Element
        for read_elem in planned_reads.findall(RunParametersXMLNodes.INNER_READ):
            read_name: str = read_elem.get(RunParametersXMLNodes.READ_NAME)
            cycles: int = int(read_elem.get(RunParametersXMLNodes.CYCLES))
            cycle_mapping[read_name] = cycles
        return cycle_mapping

    def get_index_1_cycles(self) -> int:
        """Return the number of cycles in the first index read."""
        return self.read_parser.get(RunParametersXMLNodes.INDEX_1_NOVASEQ_X)

    def get_index_2_cycles(self) -> int:
        """Return the number of cycles in the second index read."""
        return self.read_parser.get(RunParametersXMLNodes.INDEX_2_NOVASEQ_X)

    def get_read_1_cycles(self) -> int:
        """Return the number of cycles in the first read."""
        return self.read_parser.get(RunParametersXMLNodes.READ_1_NOVASEQ_X)

    def get_read_2_cycles(self) -> int:
        """Return the number of cycles in the second read."""
        return self.read_parser.get(RunParametersXMLNodes.READ_2_NOVASEQ_X)



"""Handler to delete data objects"""

from typing import List

from cg.store.filters.status_flow_cell_filters import apply_flow_cell_filter, FlowCellFilter
from cg.store.models import Flowcell, Family, Sample, SampleLaneSequencingMetrics
from cg.store.api.base import BaseHandler
from sqlalchemy.orm import Session


class DeleteDataHandler(BaseHandler):
    """Contains methods to delete business data model instances"""

    def __init__(self, session: Session):
        super().__init__(session=session)
        self.session = session

    def delete_flow_cell(self, flow_cell_id: str) -> None:
        """Delete flow cell."""
        flow_cell: Flowcell = apply_flow_cell_filter(
            flow_cells=self._get_query(table=Flowcell),
            flow_cell_name=flow_cell_id,
            filter_functions=[FlowCellFilter.GET_BY_NAME],
        ).first()

        if flow_cell:
            self.session.delete(flow_cell)
            self.session.commit()

    def delete_relationships_sample(self, sample: Sample) -> None:
        """Delete relationships between all cases and the provided sample."""
        if sample and sample.links:
            for case_sample in sample.links:
                self.session.delete(case_sample)
            self.session.commit()

    def delete_cases_without_samples(self, case_internal_ids: List[str]) -> None:
        """Delete any cases specified in case_ids without samples."""
        for case_internal_id in case_internal_ids:
            case: Family = self.get_case_by_internal_id(internal_id=case_internal_id)
            if case and not case.links:
                self.session.delete(case)
        self.session.commit()

    def delete_flow_cell_entries_in_sample_lane_sequencing_metrics(
        self, flow_cell_name: str
    ) -> None:
        """Delete all entries in sample_lane_sequencing_metrics for a flow cell."""
        metrics: List[
            SampleLaneSequencingMetrics
        ] = self.get_sample_lane_sequencing_metrics_by_flow_cell_name(flow_cell_name=flow_cell_name)
        for metric in metrics:
            if metric:
                self.session.delete(metric)
        self.session.commit()



"""Tests for the BaseHandle class."""
from cg.constants.subject import PhenotypeStatus
from sqlalchemy.orm import Query


def test_get_latest_analyses_for_cases_query(
    analysis_store, helpers, timestamp_now, timestamp_yesterday
):
    """Tests that analyses that are not latest are not returned."""

    # GIVEN an analysis a newer analysis exists
    case = helpers.add_case(analysis_store)
    analysis_oldest = helpers.add_analysis(
        analysis_store,
        case=case,
        started_at=timestamp_yesterday,
        uploaded_at=timestamp_yesterday,
        delivery_reported_at=None,
    )
    analysis_store.session.add(analysis_oldest)
    analysis_store.session.commit()
    analysis_newest = helpers.add_analysis(
        analysis_store,
        case=case,
        started_at=timestamp_now,
        uploaded_at=timestamp_now,
        delivery_reported_at=None,
    )
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp_now)
    analysis_store.relate_sample(
        family=analysis_oldest.family, sample=sample, status=PhenotypeStatus.UNKNOWN
    )

    # WHEN calling the analyses_to_delivery_report
    analyses: Query = analysis_store._get_latest_analyses_for_cases_query()

    # THEN analyses is a query
    assert isinstance(analyses, Query)

    # THEN only the newest analysis should be returned
    assert analysis_newest in analyses
    assert analysis_oldest not in analyses



from pathlib import Path
from typing import List
from cg.constants.constants import FileFormat
from cg.io.controller import ReadFile, WriteFile, ReadStream, WriteStream
from cg.models.mip.mip_sample_info import MipBaseSampleInfo


def test_get_content_from_file(case_qc_sample_info_path: Path):
    """
    Tests get_content_from_file
    """
    # GIVEN a yaml file

    # WHEN reading the yaml file
    raw_sample_info: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML, file_path=case_qc_sample_info_path
    )

    # Then assert a dict is returned
    assert isinstance(raw_sample_info, dict)

    # WHEN instantiating a MipBaseSampleInfo object
    sample_info_object = MipBaseSampleInfo(**raw_sample_info)

    # THEN assert that it was successfully created
    assert isinstance(sample_info_object, MipBaseSampleInfo)


def test_get_content_from_file_when_json(json_file_path: Path):
    """
    Tests get_content_from_file when json
    """
    # GIVEN a json file

    # WHEN reading the json file
    raw_json_content: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=json_file_path
    )

    # Then assert a dict is returned
    assert isinstance(raw_json_content, dict)


def test_get_content_from_file_when_csv(csv_file_path: Path):
    """
    Tests getting content from file when in CSV format.
    """
    # GIVEN a csv file

    # WHEN reading the csv file
    raw_csv_content: list = ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=csv_file_path
    )

    # Then assert a list is returned
    assert isinstance(raw_csv_content, list)


def test_get_content_from_file_to_dict_when_csv(csv_file_path: Path):
    """
    Tests getting content from file using DictReader when in CSV format.
    """
    # GIVEN a csv file

    # WHEN reading the csv file
    raw_csv_content: List[dict] = ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=csv_file_path, read_to_dict=True
    )

    # Then assert a list is returned and that the first element is a dict
    assert isinstance(raw_csv_content, list)
    assert isinstance(raw_csv_content[0], dict)


def test_get_content_from_stream(yaml_stream: str):
    """
    Tests read_yaml_stream
    """
    # GIVEN a string in yaml format

    # WHEN reading the yaml content in string
    raw_content: list = ReadStream.get_content_from_stream(
        file_format=FileFormat.YAML, stream=yaml_stream
    )

    # THEN assert that a list is returned
    assert isinstance(raw_content, list)


def test_get_content_from_stream_when_json(json_stream: str):
    """
    Tests read_json_stream
    """
    # GIVEN a string in json format

    # WHEN reading the json content in string
    raw_content: dict = ReadStream.get_content_from_stream(
        file_format=FileFormat.JSON, stream=json_stream
    )

    # THEN assert that a dict is returned
    assert isinstance(raw_content, dict)


def test_get_content_from_stream_when_csv(csv_stream: str):
    """
    Tests read a CVS stream.
    """
    # GIVEN a string in csv format

    # WHEN reading the csv content in string
    raw_content: list = ReadStream.get_content_from_stream(
        file_format=FileFormat.CSV, stream=csv_stream
    )

    # THEN assert that a list is returned
    assert isinstance(raw_content, list)


def test_write_file_from_content(case_qc_sample_info_path: Path, cg_dir: Path):
    """
    Tests write_file_from_content
    """
    # GIVEN a yaml file

    # GIVEN a file path to write to
    yaml_file: Path = Path(cg_dir, "write_yaml.yaml")

    # WHEN reading the yaml file
    raw_sample_info: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML, file_path=case_qc_sample_info_path
    )

    # WHEN writing the yaml file from dict
    WriteFile.write_file_from_content(
        content=raw_sample_info, file_format=FileFormat.YAML, file_path=yaml_file
    )

    # THEN assert that a file was successfully created
    assert Path.exists(yaml_file)

    # WHEN reading it as a yaml
    written_raw_sample_info: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML, file_path=yaml_file
    )

    # THEN assert that all data is kept
    assert raw_sample_info == written_raw_sample_info


def test_write_file_from_content_when_json(json_file_path: Path, json_temp_path: Path):
    """
    Tests write_file_from_content when json
    """
    # GIVEN a json file

    # GIVEN a file path to write to

    # WHEN reading the json file
    raw_json_content: list = ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=json_file_path
    )

    # WHEN writing the json file from list
    WriteFile.write_file_from_content(
        content=raw_json_content, file_format=FileFormat.JSON, file_path=json_temp_path
    )

    # THEN assert that a file was successfully created
    assert Path.exists(json_temp_path)

    # WHEN reading it as a yaml
    written_raw_sample_info: list = ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=json_temp_path
    )

    # THEN assert that all data is kept
    assert raw_json_content == written_raw_sample_info


def test_write_file_from_content_when_csv(csv_file_path: Path, csv_temp_path: Path):
    """
    Tests write file from content when CSV.
    """
    # GIVEN a csv file

    # GIVEN a file path to write to

    # WHEN reading the csv file
    raw_csv_content: list = ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=csv_file_path
    )

    # WHEN writing the csv file from list
    WriteFile.write_file_from_content(
        content=raw_csv_content, file_format=FileFormat.CSV, file_path=csv_temp_path
    )

    # THEN assert that a file was successfully created
    assert Path.exists(csv_temp_path)

    # WHEN reading it as a csv
    written_raw_sample_info: list = ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=csv_temp_path
    )

    # THEN assert that all data is kept
    assert raw_csv_content == written_raw_sample_info


def test_write_yaml_stream_from_content(yaml_stream: str):
    """
    Tests read_yaml_stream
    """
    # GIVEN a string in yaml format

    # WHEN reading the yaml content in string
    raw_content: list = ReadStream.get_content_from_stream(
        file_format=FileFormat.YAML, stream=yaml_stream
    )
    # WHEN writing a yaml stream
    yaml_content = WriteStream.write_stream_from_content(
        content=raw_content, file_format=FileFormat.YAML
    )

    # THEN assert all data is kept abd in yaml format
    assert yaml_stream == yaml_content


def test_write_json_stream_from_content(json_stream: str):
    """
    Tests read_json_stream
    """
    # GIVEN a string in json format

    # WHEN reading the json content in string
    raw_content: list = ReadStream.get_content_from_stream(
        file_format=FileFormat.JSON, stream=json_stream
    )
    # WHEN writing a json stream
    json_content = WriteStream.write_stream_from_content(
        content=raw_content, file_format=FileFormat.JSON
    )

    # THEN assert all data is kept abd in json format
    assert json_stream == json_content


def test_write_csv_stream_from_content(csv_stream: str):
    """
    Tests read CSV stream.
    """
    # GIVEN a string in csv format

    # WHEN reading the csv content in string
    raw_content: list = ReadStream.get_content_from_stream(
        file_format=FileFormat.CSV, stream=csv_stream
    )
    # WHEN writing a csv stream
    csv_content = WriteStream.write_stream_from_content(
        content=raw_content, file_format=FileFormat.CSV
    )

    # THEN assert all data is kept and in csv format
    assert csv_stream + "\n" == csv_content



"""Tests for delivery API"""

from pathlib import Path
from typing import List, Set

from cgmodels.cg.constants import Pipeline
from housekeeper.store.models import Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.delivery import INBOX_NAME
from cg.constants.housekeeper_tags import AlignmentFileTag
from cg.meta.deliver import DeliverAPI
from cg.store import Store
from cg.store.models import FamilySample, Sample, Family
from tests.store_helpers import StoreHelpers
from tests.store.conftest import fixture_case_obj
from tests.cli.deliver.conftest import fixture_fastq_delivery_bundle, fixture_mip_delivery_bundle


def test_get_delivery_path(
    base_store: Store, real_housekeeper_api: HousekeeperAPI, project_dir: Path, case_id: str
):
    """Test to create the delivery path."""
    # GIVEN a deliver api
    deliver_api = DeliverAPI(
        store=base_store,
        hk_api=real_housekeeper_api,
        case_tags=["case-tag"],
        sample_tags=["sample-tag"],
        project_base_path=project_dir,
        delivery_type="balsamic",
    )
    customer_id = "cust000"
    ticket = "1234"
    deliver_api._set_customer_id(customer_id)
    deliver_api.set_ticket(ticket)

    # WHEN fetching the deliver path
    deliver_path = deliver_api.create_delivery_dir_path(case_name=case_id)

    # THEN assert that the path looks like expected
    assert deliver_path == Path(project_dir, customer_id, INBOX_NAME, ticket, case_id)


def test_get_case_analysis_files(populated_deliver_api: DeliverAPI, case_id: str):
    """Test to fetch case specific files for a case that exists in housekeeper."""
    deliver_api: DeliverAPI = populated_deliver_api
    # GIVEN a case which exist as bundle in hk with a version
    version: Version = deliver_api.hk_api.last_version(case_id)
    assert version

    # GIVEN that a case object exists in the database
    link_objs: List[FamilySample] = deliver_api.store.get_case_samples_by_case_id(
        case_internal_id=case_id
    )
    samples: List[Sample] = [link.sample for link in link_objs]
    sample_ids: Set[str] = set([sample.internal_id for sample in samples])

    # WHEN fetching all case files from the delivery api
    bundle_latest_files = deliver_api.get_case_files_from_version(
        version=version, sample_ids=sample_ids
    )

    # THEN housekeeper files should be returned
    assert bundle_latest_files


def test_get_case_files_from_version(
    analysis_store: Store,
    case_id: str,
    real_housekeeper_api: HousekeeperAPI,
    case_hk_bundle_no_files: dict,
    bed_file: Path,
    vcf_file: Path,
    project_dir: Path,
    helpers=StoreHelpers,
):
    # GIVEN a store with a case
    case_obj = analysis_store.get_case_by_internal_id(internal_id=case_id)
    assert case_obj.internal_id == case_id
    # GIVEN a delivery api
    deliver_api = DeliverAPI(
        store=analysis_store,
        hk_api=real_housekeeper_api,
        case_tags=[{"case-tag"}],
        sample_tags=[{"sample-tag"}],
        project_base_path=project_dir,
        delivery_type="balsamic",
    )

    # GIVEN a housekeeper db populated with a bundle including a case specific file and a sample specific file
    case_hk_bundle_no_files["files"] = [
        {"path": bed_file.as_posix(), "archive": False, "tags": ["case-tag"]},
        {"path": str(vcf_file), "archive": False, "tags": ["sample-tag", "ADM1"]},
    ]
    helpers.ensure_hk_bundle(real_housekeeper_api, bundle_data=case_hk_bundle_no_files)

    # GIVEN a version object where two file exists
    version: Version = real_housekeeper_api.last_version(bundle=case_id)
    assert len(version.files) == 2

    # GIVEN the sample ids of the samples
    link_objs: List[FamilySample] = analysis_store.get_case_samples_by_case_id(case_id)
    samples: List[Sample] = [link.sample for link in link_objs]
    sample_ids: Set[str] = set([sample.internal_id for sample in samples])

    # WHEN fetching the case files
    case_files = deliver_api.get_case_files_from_version(version=version, sample_ids=sample_ids)

    # THEN we should only get the case specific files back
    nr_files: int = 0
    case_file: Path
    for nr_files, case_file in enumerate(case_files, 1):
        assert case_file.name == bed_file.name
    # THEN assert that only the case-tag file was returned
    assert nr_files == 1


def test_get_sample_files_from_version(
    deliver_api: DeliverAPI,
    case_hk_bundle_no_files: dict,
    bed_file: Path,
    vcf_file: Path,
    project_dir: Path,
    case_id: str,
    helpers=StoreHelpers,
):
    """Test to fetch sample specific files from the deliver API.

    The purpose of the test is to see that only sample specific files are returned.
    """
    # GIVEN a case which exist as bundle in hk
    # GIVEN a housekeeper db populated with a bundle including a case specific file and a sample specific file
    hk_api = deliver_api.hk_api
    case_hk_bundle_no_files["files"] = [
        {"path": bed_file.as_posix(), "archive": False, "tags": ["case-tag"]},
        {"path": str(vcf_file), "archive": False, "tags": [AlignmentFileTag.CRAM, "ADM1"]},
    ]
    helpers.ensure_hk_bundle(hk_api, bundle_data=case_hk_bundle_no_files)
    # GIVEN a version object with some files
    version: Version = hk_api.last_version(case_id)
    assert len(version.files) == 2

    # WHEN fetching the sample specific files
    sample_files = deliver_api.get_sample_files_from_version(version_obj=version, sample_id="ADM1")

    # THEN assert that only the sample specific file was returned
    nr_files: int = 0
    sample_file: Path
    for nr_files, sample_file in enumerate(sample_files, 1):
        assert sample_file.name == vcf_file.name
    # THEN assert that only the sample-tag file was returned
    assert nr_files == 1


def test_get_delivery_scope_case_only():
    """Testing the delivery scope of a case only delivery."""
    # GIVEN a case only delivery type
    delivery_type: Set[str] = {Pipeline.MIP_DNA}

    # WHEN getting the delivery scope
    sample_delivery, case_delivery = DeliverAPI.get_delivery_scope(delivery_type)

    # THEN a case_delivery should be True while sample_delivery False
    assert case_delivery
    assert not sample_delivery


def test_get_delivery_scope_sample_only():
    """Testing the delivery scope of a sample only delivery."""
    # GIVEN a sample only delivery type
    delivery_type = {Pipeline.FASTQ}

    # WHEN getting the delivery scope
    sample_delivery, case_delivery = DeliverAPI.get_delivery_scope(delivery_type)

    # THEN a sample_delivery should be True while case_delivery False
    assert not case_delivery
    assert sample_delivery


def test_get_delivery_scope_case_and_sample():
    """Testing the delivery scope of a case and sample delivery."""
    # GIVEN a case and sample delivery type
    delivery_type = {Pipeline.SARS_COV_2}

    # WHEN getting the delivery scope
    sample_delivery, case_delivery = DeliverAPI.get_delivery_scope(delivery_type)

    # THEN both case_delivery and sample_delivery should be True
    assert case_delivery
    assert sample_delivery


def test_deliver_files_enough_reads(
    caplog,
    case_id: str,
    deliver_api: DeliverAPI,
    deliver_api_destination_path: Path,
    fastq_delivery_bundle: dict,
    helpers: StoreHelpers,
    mip_delivery_bundle: dict,
    sample_id: str,
):
    """Tests the deliver_files method for a sample with enough reads."""
    # GIVEN a case to be delivered and a sample with enough reads
    case: Family = deliver_api.store.get_case_by_internal_id(internal_id=case_id)
    sample: Sample = deliver_api.store.get_sample_by_internal_id(sample_id)
    helpers.ensure_hk_bundle(deliver_api.hk_api, fastq_delivery_bundle, include=True)
    helpers.ensure_hk_bundle(deliver_api.hk_api, mip_delivery_bundle, include=True)

    # WHEN delivering files for the case
    deliver_api.deliver_files(case_obj=case)

    # THEN the sample folder should be created
    assert Path(deliver_api.project_base_path, deliver_api_destination_path, sample.name).exists()


def test_deliver_files_not_enough_reads(
    caplog,
    case_id: str,
    deliver_api: DeliverAPI,
    deliver_api_destination_path: Path,
    fastq_delivery_bundle: dict,
    helpers: StoreHelpers,
    mip_delivery_bundle: dict,
    sample_id: str,
):
    """Tests the deliver_files method for a sample with too few reads."""
    # GIVEN a case to be delivered and a sample with too few reads
    case: Family = deliver_api.store.get_case_by_internal_id(internal_id=case_id)
    sample: Sample = deliver_api.store.get_sample_by_internal_id(sample_id)
    sample.reads = 1
    helpers.ensure_hk_bundle(deliver_api.hk_api, fastq_delivery_bundle, include=True)
    helpers.ensure_hk_bundle(deliver_api.hk_api, mip_delivery_bundle, include=True)

    # WHEN delivering files for the case
    deliver_api.deliver_files(case_obj=case)

    # THEN the sample folder should not be created
    assert not Path(
        deliver_api.project_base_path, deliver_api_destination_path, sample.name
    ).exists()


def test_deliver_files_not_enough_reads_force(
    caplog,
    case_id: str,
    deliver_api: DeliverAPI,
    deliver_api_destination_path: Path,
    fastq_delivery_bundle: dict,
    helpers: StoreHelpers,
    mip_delivery_bundle: dict,
    sample_id: str,
):
    """Tests the deliver_files method for a sample with too few reads but with override."""
    # GIVEN a case to be delivered and a sample with too few reads
    case: Family = deliver_api.store.get_case_by_internal_id(internal_id=case_id)
    sample: Sample = deliver_api.store.get_sample_by_internal_id(sample_id)
    sample.reads = 1
    helpers.ensure_hk_bundle(deliver_api.hk_api, fastq_delivery_bundle, include=True)
    helpers.ensure_hk_bundle(deliver_api.hk_api, mip_delivery_bundle, include=True)

    # Given that the API was created with force_all=True
    deliver_api.deliver_failed_samples = True

    # WHEN delivering files for the case
    deliver_api.deliver_files(
        case_obj=case,
    )

    # THEN the sample folder should be created
    assert Path(deliver_api.project_base_path, deliver_api_destination_path, sample.name).exists()



from pydantic.v1 import BaseModel, validator


class MutantSampleConfig(BaseModel):
    CG_ID_project: str
    CG_ID_sample: str
    case_ID: str
    region_code: str
    lab_code: str
    priority: str
    Customer_ID_project: int
    Customer_ID_sample: str
    customer_id: str
    application_tag: str
    date_arrival: str
    date_libprep: str
    date_sequencing: str
    method_libprep: str
    method_sequencing: str
    sequencing_qc_pass: bool
    selection_criteria: str
    primer: str

    @validator("region_code", "lab_code")
    def sanitize_values(cls, value):
        return value.split(" ")[0]



"""Maps tag info from housekeeper tags to scout load config"""

from typing import Optional, Set

from pydantic.v1 import BaseModel, Field


class CaseTags(BaseModel):
    snv_vcf: Optional[Set[str]] = Field(
        None, description="vcf_snv for rare disease and vcf_cancer for cancer"
    )
    snv_research_vcf: Optional[Set[str]] = Field(
        None, description="vcf_snv_research for rare disease"
    )
    sv_vcf: Optional[Set[str]] = Field(
        None, description="vcf_cancer_sv for rare disease and vcf_sv_cancer for cancer"
    )
    sv_research_vcf: Optional[Set[str]] = Field(
        None, description="vcf_sv_research for rare disease"
    )
    vcf_str: Set[str] = Field(
        None, description="Short Tandem Repeat variants, only for rare disease"
    )
    cnv_report: Optional[Set[str]] = Field(
        None, description="AscatNgs visualization report for cancer"
    )
    smn_tsv: Optional[Set[str]] = Field(
        None, description="SMN gene variants, only for rare disease"
    )
    peddy_ped: Set[str] = Field(None, description="Ped info from peddy, only for rare disease")
    peddy_sex: Optional[Set[str]] = Field(
        None, description="Peddy sex check, only for rare disease"
    )
    peddy_check: Set[str] = Field(None, description="Peddy pedigree check, only for rare disease")
    multiqc_report: Optional[Set[str]] = Field(None, description="MultiQC report")
    delivery_report: Optional[Set[str]] = Field(None, description="Delivery report")
    str_catalog: Optional[Set[str]] = Field(
        None, description="Variant catalog used with expansionhunter"
    )
    gene_fusion: Set[str] = Field(
        None, description="Arriba report for RNA fusions containing only clinical fusions"
    )
    gene_fusion_report_research: Optional[Set[str]] = Field(
        None, description="Arriba report for RNA fusions containing all fusions"
    )
    RNAfusion_report: Optional[Set[str]] = Field(
        None, description="Main RNA fusion report containing only clinical fusions"
    )
    RNAfusion_report_research: Optional[Set[str]] = Field(
        None, description="Main RNA fusion report containing all fusions"
    )
    RNAfusion_inspector: Optional[Set[str]] = Field(
        None, description="RNAfusion inspector report containing only clinical fusions"
    )
    RNAfusion_inspector_research: Optional[Set[str]] = Field(
        None, description="RNAfusion inspector report containing all fusions"
    )
    multiqc_rna: Optional[Set[str]] = Field(None, description="MultiQC report for RNA samples")
    vcf_mei: Optional[Set[str]] = Field(
        None, description="VCF with mobile element insertions, clinical"
    )
    vcf_mei_research: Optional[Set[str]] = Field(
        None, description="VCF with mobile element insertions, research"
    )


class SampleTags(BaseModel):
    # If cram does not exist
    bam_file: Optional[Set[str]] = None
    alignment_file: Optional[Set[str]] = None
    vcf2cytosure: Optional[Set[str]] = None
    mt_bam: Optional[Set[str]] = None
    chromograph_autozyg: Optional[Set[str]] = None
    chromograph_coverage: Optional[Set[str]] = None
    chromograph_regions: Optional[Set[str]] = None
    chromograph_sites: Optional[Set[str]] = None
    reviewer_alignment: Optional[Set[str]] = None
    reviewer_alignment_index: Optional[Set[str]] = None
    reviewer_vcf: Optional[Set[str]] = None
    mitodel_file: Optional[Set[str]] = None



from cg.apps.coverage import ChanjoAPI
from cg.apps.crunchy import CrunchyAPI
from cg.apps.gt import GenotypeAPI
from cg.apps.hermes.hermes_api import HermesApi
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.madeline.api import MadelineAPI
from cg.apps.mutacc_auto import MutaccAutoAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.apps.tb import TrailblazerAPI
from cg.meta.backup.backup import SpringBackupAPI
from cg.meta.compress import CompressAPI
from cg.meta.encryption.encryption import SpringEncryptionAPI
from cg.meta.workflow.prepare_fastq import PrepareFastqAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.meta.backup.pdc import PdcAPI


class MetaAPI:

    """MetaAPI class initializing all App APIs used within CG in non-conflicting manner"""

    def __init__(self, config: CGConfig):
        self.config = config
        self.chanjo_api: ChanjoAPI = config.chanjo_api
        self.crunchy_api: CrunchyAPI = config.crunchy_api
        self.genotype_api: GenotypeAPI = config.genotype_api
        self.hermes_api: HermesApi = config.hermes_api
        self.housekeeper_api: HousekeeperAPI = config.housekeeper_api
        self.lims_api: LimsAPI = config.lims_api
        self.madeline_api: MadelineAPI = config.madeline_api
        self.mutacc_auto_api: MutaccAutoAPI = config.mutacc_auto_api
        self.prepare_fastq_api: PrepareFastqAPI = PrepareFastqAPI(
            store=config.status_db,
            compress_api=CompressAPI(
                hk_api=config.housekeeper_api,
                crunchy_api=config.crunchy_api,
                demux_root=config.demultiplex.out_dir,
                backup_api=SpringBackupAPI(
                    encryption_api=SpringEncryptionAPI(
                        binary_path=config.dict()["encryption"]["binary_path"]
                    ),
                    hk_api=config.housekeeper_api,
                    pdc_api=PdcAPI(config.dict()["pdc"]["binary_path"]),
                ),
            ),
        )
        self.status_db: Store = config.status_db
        self.scout_api: ScoutAPI = config.scout_api
        self.trailblazer_api: TrailblazerAPI = config.trailblazer_api



"""Get samples with information from Dragen demultiplexing"""


import logging

from pydantic.v1 import BaseModel

LOG = logging.getLogger(__name__)


class DragenDemuxSample(BaseModel):
    """Gather statistics from the Dragen demultiplexing results for a sample"""

    sample_name: str
    flowcell: str
    lane: int
    reads: int = 0
    perfect_reads: int = 0
    one_mismatch_reads: int = 0
    pass_filter_q30: int = 0
    mean_quality_score: float = 0.00
    r1_sample_bases: int = 0
    r2_sample_bases: int = 0
    read_length: int = 0



import logging

from cg.cli.workflow.balsamic.base import link
from cg.models.cg_config import CGConfig
from click.testing import CliRunner

EXIT_SUCCESS = 0


def test_without_options(cli_runner: CliRunner, balsamic_context: CGConfig):
    """Test command without case_id"""
    # GIVEN NO case_id
    # WHEN dry running without anything specified
    result = cli_runner.invoke(link, obj=balsamic_context)
    # THEN command should mention argument
    assert result.exit_code != EXIT_SUCCESS
    assert "Missing argument" in result.output


def test_with_missing_case(cli_runner: CliRunner, balsamic_context: CGConfig, caplog):
    """Test command with invalid case to start with"""
    caplog.set_level(logging.ERROR)
    # GIVEN case_id not in database
    case_id = "soberelephant"
    assert not balsamic_context.status_db.get_case_by_internal_id(internal_id=case_id)
    # WHEN running
    result = cli_runner.invoke(link, [case_id], obj=balsamic_context)
    # THEN command should NOT successfully call the command it creates
    assert result.exit_code != EXIT_SUCCESS
    # THEN ERROR log should be printed containing invalid case_id
    assert "could not be found in Status DB!" in caplog.text


def test_without_samples(cli_runner: CliRunner, balsamic_context: CGConfig, caplog):
    """Test command with case_id and no samples"""
    caplog.set_level(logging.ERROR)
    # GIVEN case-id
    case_id = "no_sample_case"
    # WHEN dry running with dry specified
    result = cli_runner.invoke(link, [case_id], obj=balsamic_context)
    # THEN command should print the balsamic command-string
    assert result.exit_code != EXIT_SUCCESS
    # THEN warning should be printed that no samples are found
    assert case_id in caplog.text
    assert "no samples" in caplog.text


def test_single_panel(cli_runner: CliRunner, balsamic_context: CGConfig, caplog):
    """Test with case_id that requires SINGLE TGS analysis"""
    caplog.set_level(logging.INFO)
    # GIVEN case_id containing ONE tumor, TGS application
    case_id = "balsamic_case_tgs_single"
    # WHEN dry running
    result = cli_runner.invoke(link, [case_id], obj=balsamic_context)
    # THEN command is generated successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN log should inform us that files are linked
    assert "Linking" in caplog.text



from typing import List, Dict
from pydantic.v1 import BaseModel, Field, validator


class IndexMetric(BaseModel):
    """
    Represents a metric generated by bcl2fastq for a specific index sequence in the demultiplexing process.
    Each instance of this class includes the index sequence itself and a dictionary
    mapping mismatch counts to their frequencies.

    Fields:
    - mismatch_counts: A dictionary mapping mismatch counts (as strings) to their
      respective frequencies.
    """

    mismatch_counts: Dict[str, int] = Field(..., alias="MismatchCounts")

    @validator("mismatch_counts", each_item=True)
    def check_non_negative(cls, value):
        if value < 0:
            raise ValueError("MismatchCounts must be non-negative")
        return value


class ReadMetric(BaseModel):
    """
    Represents a set of metrics generated by bcl2fastq for a read in the sequencing process.

    Fields:
    - read_number: The read number this metric pertains to.
    - yield_: The yield from this read.
    - yield_q30: The yield with quality score greater than or equal to 30 from this read.
    - quality_score_sum: The sum of quality scores from this read.
    """

    read_number: int = Field(..., alias="ReadNumber", gt=0)
    yield_: int = Field(..., alias="Yield", ge=0)
    yield_q30: int = Field(..., alias="YieldQ30", ge=0)
    quality_score_sum: int = Field(..., alias="QualityScoreSum", ge=0)


class DemuxResult(BaseModel):
    """
    Represents the result of the bcl2fastq demultiplexing process for a given sample.

    Fields:
    - sample_id: The unique identifier for the sample.
    - sample_name: The name of the sample.
    - index_metrics: A list of metrics for each index sequence related to this sample.
    - number_reads: The total number of reads for this sample.
    - yield_: The total yield from this sample.
    - read_metrics: A list of read metrics for each read related to this sample.
    """

    sample_id: str = Field(..., alias="SampleId", min_length=1)
    sample_name: str = Field(..., alias="SampleName", min_length=1)
    index_metrics: List[IndexMetric] = Field(..., alias="IndexMetrics")
    number_reads: int = Field(..., alias="NumberReads", gt=0)
    yield_: int = Field(..., alias="Yield", ge=0)
    read_metrics: List[ReadMetric] = Field(..., alias="ReadMetrics")


class ConversionResult(BaseModel):
    """
    Represents the result of the conversion process for a given lane as generated by
    the bcl2fastq demultiplexing software.

    Fields:
    - lane_number: The number of the lane this result pertains to.
    - yield_: The total yield from this lane.
    - demux_results: A list of demultiplexing results for each sample in this lane.
    """

    lane_number: int = Field(..., alias="LaneNumber", gt=0)
    yield_: int = Field(..., alias="Yield", ge=0)
    demux_results: List[DemuxResult] = Field(..., alias="DemuxResults")


class ReadInfoForLane(BaseModel):
    """
    Represents basic information for a given lane.

    Fields:
    - lane_number: The number of the lane.
    """

    lane_number: int = Field(..., alias="LaneNumber", gt=0)


class Bcl2FastqSampleLaneTileMetrics(BaseModel):
    """
    Represents a set of sequencing metrics for a bcl2fastq run per sample, lane and tile on a flow cell.

    Fields:
    - flowcell: The identifier for the flowcell used in this run.
    - read_infos_for_lanes: A list of basic information for each lane in this run.
    - conversion_results: A list of conversion results for each lane in this run.
    """

    flow_cell_name: str = Field(..., alias="Flowcell", min_length=1)
    read_infos_for_lanes: List[ReadInfoForLane] = Field(..., alias="ReadInfosForLanes")
    conversion_results: List[ConversionResult] = Field(..., alias="ConversionResults")


class Bcl2FastqSampleLaneMetrics(BaseModel):
    """
    Aggregated metrics per sample and lane from a bcl2fastq run.

     Attributes:
     - flow_cell_name (str): Name of the flow cell.
     - flow_cell_lane_number (int): Lane number within the flow cell.
     - sample_id (str): Unique identifier of the sample.
     - sample_total_reads_in_lane (int): Total number of reads for the sample in the lane.
     - sample_total_yield_in_lane (int): Total yield of bases for the sample in the lane.
     - sample_total_yield_q30_in_lane (int): Total yield of bases with Q30 quality score in the lane.
     - sample_total_quality_score_in_lane (int): Sum of quality scores for all bases in the sample reads in the lane.
    """

    flow_cell_name: str
    flow_cell_lane_number: int
    sample_id: str
    sample_total_reads_in_lane: int
    sample_total_yield_in_lane: int
    sample_total_yield_q30_in_lane: int
    sample_total_quality_score_in_lane: int



"""Helper functions."""

from pathlib import Path
from typing import List, Optional


def get_string_from_list_by_pattern(strings: List[str], pattern: str) -> Optional[str]:
    """Returns the full string from a list given a specific pattern."""

    return next((string for string in strings if pattern in string), None) if strings else None


def build_command_from_dict(options: dict, exclude_true: bool = False) -> List[str]:
    """Returns a command list of strings given a dictionary with arguments."""
    formatted_options: list = []
    for option_name, option_value in options.items():
        if option_value:
            if exclude_true and option_value is True:
                formatted_options.append(str(option_name))
            elif option_value:
                formatted_options.append(str(option_name))
                formatted_options.append(str(option_value))
    return formatted_options


def get_hamming_distance(str_1: str, str_2: str) -> int:
    """Return the hamming distance between two strings.

    Raises:
        KeyError: When the strings have the same length.
    """
    if len(str_1) != len(str_2):
        raise KeyError("The two strings must have the same length to calculate distance!")
    return sum(n1 != n2 for n1, n2 in zip(str_1, str_2))



import logging
import os
from pathlib import Path
from typing import Dict, Generator, List

import pytest


from cg.constants.demultiplexing import BclConverter, DemultiplexingDirsAndFiles
from cg.constants.housekeeper_tags import SequencingFileTag
from cg.meta.demultiplex.demux_post_processing import (
    DemuxPostProcessingAPI,
    DemuxPostProcessingHiseqXAPI,
)
from cg.meta.transfer import TransferFlowCell
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.demux_results import DemuxResults
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData
from cg.store import Store

from tests.meta.demultiplex.conftest import FlowCellInfo


def test_set_dry_run(
    demultiplex_context: CGConfig,
):
    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingAPI = DemuxPostProcessingAPI(config=demultiplex_context)

    # THEN dry run should be False
    assert post_demux_api.dry_run is False

    # WHEN dry run set to True
    post_demux_api.set_dry_run(dry_run=True)

    # THEN dry run should be True
    assert post_demux_api.dry_run is True


def test_add_to_cgstats_dry_run(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
):
    caplog.set_level(logging.INFO)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN dry run set to True
    post_demux_api.set_dry_run(dry_run=True)

    # When adding to cgstats
    post_demux_api.add_to_cgstats(flow_cell_path=bcl2fastq_flow_cell.path)

    # THEN we should just log and exit
    assert "Dry run will not add flow cell stats" in caplog.text


def test_add_to_cgstats(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
):
    caplog.set_level(logging.INFO)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # When adding to cgstats
    post_demux_api.add_to_cgstats(flow_cell_path=bcl2fastq_flow_cell.path)

    # THEN we should run the command
    assert f"add --machine X -u Unaligned {bcl2fastq_flow_cell.path}" in caplog.text


def test_cgstats_select_project_dry_run(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    flow_cell_project_id: int,
    cgstats_select_project_log_file: Path,
):
    caplog.set_level(logging.INFO)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN an unaligned project directory
    Path(bcl2fastq_flow_cell.path, "Unaligned", f"Project_{flow_cell_project_id}").mkdir(
        parents=True, exist_ok=True
    )

    # GIVEN dry run set to True
    post_demux_api.set_dry_run(dry_run=True)

    # When processing project with cgstats
    post_demux_api.cgstats_select_project(
        flow_cell_id=bcl2fastq_flow_cell.id, flow_cell_path=bcl2fastq_flow_cell.path
    )

    # THEN we should just log and exit
    assert "Dry run will not process selected project" in caplog.text


def test_cgstats_select_project(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    flow_cell_project_id: int,
    cgstats_select_project_log_file: Path,
):
    caplog.set_level(logging.INFO)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN an unaligned project directory
    Path(bcl2fastq_flow_cell.path, "Unaligned", f"Project_{flow_cell_project_id}").mkdir(
        parents=True, exist_ok=True
    )

    # When processing project with cgstats
    post_demux_api.cgstats_select_project(
        flow_cell_id=bcl2fastq_flow_cell.id, flow_cell_path=bcl2fastq_flow_cell.path
    )

    # THEN we should have created a stats outfile
    assert cgstats_select_project_log_file.exists()

    # Clean up from calling cgstats_select_project
    cgstats_select_project_log_file.unlink()

    # THEN we should run the command
    assert f"select --project {flow_cell_project_id} {bcl2fastq_flow_cell.id}" in caplog.text


def test_cgstats_lanestats_dry_run(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
):
    caplog.set_level(logging.INFO)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN dry run set to True
    post_demux_api.set_dry_run(dry_run=True)

    # When processing lane stats with cgstats
    post_demux_api.cgstats_lanestats(flow_cell_path=bcl2fastq_flow_cell.path)

    # THEN we should run the command
    assert "Dry run will not add lane stats" in caplog.text


def test_cgstats_lanestats(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
):
    caplog.set_level(logging.INFO)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # When processing lane stats with cgstats
    post_demux_api.cgstats_lanestats(flow_cell_path=bcl2fastq_flow_cell.path)

    # THEN we should run the command
    assert f"lanestats {bcl2fastq_flow_cell.path}" in caplog.text


def test_finish_flow_cell_copy_not_completed(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    hiseq_x_copy_complete_file: Path,
):
    caplog.set_level(logging.DEBUG)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN a not completely copied flow cell
    if hiseq_x_copy_complete_file.exists():
        hiseq_x_copy_complete_file.unlink()

    # WHEN finishing flow cell
    post_demux_api.finish_flow_cell(
        bcl_converter=BclConverter.BCL2FASTQ,
        flow_cell_name=bcl2fastq_flow_cell.full_name,
        flow_cell_path=bcl2fastq_flow_cell.path,
    )

    # Reinstate
    hiseq_x_copy_complete_file.touch()

    # THEN we should log that copy is not complete
    assert f"{bcl2fastq_flow_cell.full_name} is not yet completely copied" in caplog.text


def test_finish_flow_cell_delivery_started(
    caplog,
    demultiplexing_delivery_file: Path,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
):
    caplog.set_level(logging.DEBUG)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN an already started flag file
    demultiplexing_delivery_file.touch()

    # WHEN finishing flow cell
    post_demux_api.finish_flow_cell(
        bcl_converter=BclConverter.BCL2FASTQ,
        flow_cell_name=bcl2fastq_flow_cell.full_name,
        flow_cell_path=bcl2fastq_flow_cell.path,
    )

    # Clean up
    demultiplexing_delivery_file.unlink()

    # THEN we should log that the delivery has already started
    assert (
        f"{bcl2fastq_flow_cell.full_name} copy is complete and delivery has already started"
        in caplog.text
    )


def test_finish_flow_cell_delivery_not_hiseq_x(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    hiseq_x_tile_dir: Path,
):
    caplog.set_level(logging.DEBUG)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN no hiseq X flow cell
    if hiseq_x_tile_dir.exists():
        hiseq_x_tile_dir.rmdir()

    # WHEN finishing flow cell
    post_demux_api.finish_flow_cell(
        bcl_converter=BclConverter.BCL2FASTQ,
        flow_cell_name=bcl2fastq_flow_cell.full_name,
        flow_cell_path=bcl2fastq_flow_cell.path,
    )

    # THEN we should log that this is not an Hiseq X flow cell
    assert f"{bcl2fastq_flow_cell.full_name} is not an Hiseq X flow cell" in caplog.text


def test_finish_flow_cell_ready(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    flow_cell_project_id: int,
    demultiplexing_delivery_file: Path,
    hiseq_x_tile_dir: Path,
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None],
):
    caplog.set_level(logging.INFO)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN dry run set to True
    post_demux_api.set_dry_run(dry_run=True)

    # GIVEN a Hiseq X tile directory
    hiseq_x_tile_dir.mkdir(parents=True, exist_ok=True)

    # GIVEN an unaligned project directory
    Path(
        bcl2fastq_flow_cell.path,
        DemultiplexingDirsAndFiles.UNALIGNED_DIR_NAME,
        f"Project_{flow_cell_project_id}",
    ).mkdir(parents=True, exist_ok=True)

    # WHEN finishing flow cell
    post_demux_api.finish_flow_cell(
        bcl_converter=BclConverter.BCL2FASTQ,
        flow_cell_name=bcl2fastq_flow_cell.full_name,
        flow_cell_path=bcl2fastq_flow_cell.path,
    )

    # THEN we should log that post-processing will begin
    assert (
        f"{bcl2fastq_flow_cell.full_name} copy is complete and delivery will start" in caplog.text
    )


def test_post_process_flow_cell_dry_run(
    bcl2fastq_demux_results: DemuxResults,
    caplog,
    demultiplexing_delivery_file: Path,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    flow_cell_project_id: int,
    flowcell_store: Store,
    hiseq_x_tile_dir: Path,
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None],
):
    caplog.set_level(logging.DEBUG)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN a Hiseq X tile directory
    hiseq_x_tile_dir.mkdir(parents=True, exist_ok=True)

    # GIVEN an unaligned project directory
    Path(
        bcl2fastq_flow_cell.path,
        DemultiplexingDirsAndFiles.UNALIGNED_DIR_NAME,
        f"Project_{flow_cell_project_id}",
    ).mkdir(parents=True, exist_ok=True)

    # GIVEN dry run set to True
    post_demux_api.set_dry_run(dry_run=True)

    # WHEN post-processing flow cell
    post_demux_api.post_process_flow_cell(demux_results=bcl2fastq_demux_results)

    # THEN a delivery file should not have been created
    assert not demultiplexing_delivery_file.exists()

    # THEN we should log that we will not commit
    assert "Dry run will not commit flow cell to database" in caplog.text


def test_post_process_flow_cell(
    bcl2fastq_demux_results: DemuxResults,
    caplog,
    cgstats_select_project_log_file: Path,
    demultiplexing_delivery_file: Path,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    flow_cell_project_id: int,
    hiseq_x_tile_dir: Path,
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None],
):
    caplog.set_level(logging.DEBUG)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN a Hiseq X tile directory
    hiseq_x_tile_dir.mkdir(parents=True, exist_ok=True)

    # GIVEN an unaligned project directory
    Path(
        bcl2fastq_flow_cell.path,
        DemultiplexingDirsAndFiles.UNALIGNED_DIR_NAME,
        f"Project_{flow_cell_project_id}",
    ).mkdir(parents=True, exist_ok=True)

    # WHEN post-processing flow cell
    post_demux_api.post_process_flow_cell(demux_results=bcl2fastq_demux_results)

    # THEN a delivery file should have been created
    assert demultiplexing_delivery_file.exists()

    # Clean up
    cgstats_select_project_log_file.unlink()
    demultiplexing_delivery_file.unlink()

    # THEN we should also transfer the flow cell
    assert f"Flow cell added: {bcl2fastq_flow_cell.id}" in caplog.text


def test_finish_flow_cell(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    hiseq_x_copy_complete_file: Path,
):
    caplog.set_level(logging.DEBUG)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN a not completely copied flow cell
    hiseq_x_copy_complete_file.unlink()

    # When post-processing flow cell
    post_demux_api.finish_flow_cell(
        bcl_converter=BclConverter.BCL2FASTQ,
        flow_cell_name=bcl2fastq_flow_cell.full_name,
        flow_cell_path=bcl2fastq_flow_cell.path,
    )

    # Reinstate
    hiseq_x_copy_complete_file.touch()

    # THEN we should log that we are checking flow cell
    assert f"Check demultiplexed flow cell {bcl2fastq_flow_cell.full_name}" in caplog.text


def test_finish_all_flowcells(
    caplog,
    tmp_demultiplexed_runs_directory: Path,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    hiseq_x_copy_complete_file: Path,
):
    caplog.set_level(logging.DEBUG)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN a not completely copied flow cell
    hiseq_x_copy_complete_file.unlink()

    # When post-processing flow cell
    post_demux_api.finish_all_flow_cells(
        bcl_converter=BclConverter.BCL2FASTQ,
    )

    # Reinstate
    hiseq_x_copy_complete_file.touch()

    # THEN we should log that we are checking flow cell
    assert f"Check demultiplexed flow cell {bcl2fastq_flow_cell.full_name}" in caplog.text


@pytest.mark.parametrize(
    "demux_type",
    ["BCL2FASTQ_TREE", "BCLCONVERT_TREE", "BCLCONVERT_FLAT"],
)
def test_post_processing_of_flow_cell(
    demux_type: str,
    demultiplex_context: CGConfig,
    flow_cell_info_map: Dict[str, FlowCellInfo],
    tmp_demultiplexed_runs_directory: Path,
):
    """Test adding a demultiplexed flow cell to the databases with. Runs on each type of
    demultiplexing software and setting used."""

    # GIVEN a demultiplexed flow cell
    flow_cell_demultplexing_directory: str = flow_cell_info_map.get(demux_type).directory
    flow_cell_name: str = flow_cell_info_map.get(demux_type).name
    sample_internal_ids: List[str] = flow_cell_info_map.get(demux_type).sample_internal_ids

    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    # GIVEN a directory with a flow cell demultiplexed with BCL Convert
    demux_post_processing_api.demux_api.demultiplexed_runs_dir = tmp_demultiplexed_runs_directory

    # GIVEN that a sample sheet exists in the flow cell run directory
    path = Path(
        demux_post_processing_api.demux_api.flow_cells_dir,
        flow_cell_demultplexing_directory,
        DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME,
    )
    os.makedirs(path.parent, exist_ok=True)
    path.touch()

    # WHEN post processing the demultiplexed flow cell
    demux_post_processing_api.finish_flow_cell_temp(flow_cell_demultplexing_directory)

    # THEN a flow cell was created in statusdb
    assert demux_post_processing_api.status_db.get_flow_cell_by_name(flow_cell_name)

    # THEN sequencing metrics were created for the flow cell
    assert demux_post_processing_api.status_db.get_sample_lane_sequencing_metrics_by_flow_cell_name(
        flow_cell_name=flow_cell_name
    )
    # THEN the read count was calculated for all samples in the flow cell directory
    for sample_internal_id in sample_internal_ids:
        sample = demux_post_processing_api.status_db.get_sample_by_internal_id(sample_internal_id)
        assert sample is not None
        assert sample.reads

    # THEN a bundle was added to Housekeeper for the flow cell
    assert demux_post_processing_api.hk_api.bundle(flow_cell_name)

    # THEN a bundle was added to Housekeeper for each sample
    for sample_internal_id in sample_internal_ids:
        assert demux_post_processing_api.hk_api.bundle(sample_internal_id)

    # THEN a sample sheet was added to Housekeeper
    assert demux_post_processing_api.hk_api.get_files(
        tags=[SequencingFileTag.SAMPLE_SHEET],
        bundle=flow_cell_name,
    ).all()

    # THEN sample fastq files were added to Housekeeper tagged with FASTQ and the flow cell name
    for sample_internal_id in sample_internal_ids:
        assert demux_post_processing_api.hk_api.get_files(
            tags=[SequencingFileTag.FASTQ, flow_cell_name],
            bundle=sample_internal_id,
        ).all()

    # THEN a delivery file was created in the flow cell directory
    delivery_path = Path(
        demux_post_processing_api.demux_api.demultiplexed_runs_dir,
        flow_cell_demultplexing_directory,
        DemultiplexingDirsAndFiles.DELIVERY,
    )

    assert delivery_path.exists()


def test_get_all_demultiplexed_flow_cell_out_dirs(
    demultiplex_context: CGConfig,
    tmp_demultiplexed_runs_directory: Path,
    tmp_demultiplexed_runs_bcl2fastq_directory: Path,
):
    """Test returning all flow cell directories from the demultiplexing run directory."""
    # GIVEN a demultiplex flow cell finished output directory that exist

    # GIVEN a demultiplex context
    demux_api: DemuxPostProcessingAPI = DemuxPostProcessingAPI(config=demultiplex_context)

    # WHEN calling get_all_demultiplexed_flow_cell_dirs
    demultiplex_flow_cell_dirs: List[Path] = demux_api.get_all_demultiplexed_flow_cell_dirs()

    # THEN the demultiplexed flow cells run directories should be returned
    assert tmp_demultiplexed_runs_bcl2fastq_directory in demultiplex_flow_cell_dirs



from cg.utils.files import get_file_in_directory
from pathlib import Path


def test_get_file_in_directory(nested_directory_with_file: Path, some_file: str):
    """Test function to get a file in a directory and subdirectories."""
    # GIVEN a directory with subdirectories with a file
    # WHEN getting the file
    file_path: Path = get_file_in_directory(nested_directory_with_file, some_file)
    # THEN assert that the file is returned
    assert file_path.exists()



"""RNAfusion related constants."""

from cg.constants.nextflow import NFX_SAMPLESHEET_HEADERS

RNAFUSION_STRANDEDNESS_HEADER = "strandedness"
RNAFUSION_SAMPLESHEET_HEADERS = NFX_SAMPLESHEET_HEADERS + [RNAFUSION_STRANDEDNESS_HEADER]
RNAFUSION_METRIC_CONDITIONS: dict = {
    "uniquely_mapped_percent": {"norm": "gt", "threshold": 60},
    "PCT_MRNA_BASES": {"norm": "gt", "threshold": 80},
    "PCT_RIBOSOMAL_BASES": {"norm": "lt", "threshold": 5},
    "PERCENT_DUPLICATION": {"norm": "lt", "threshold": 0.7},
}



"""
Parse statistics from the demultiplexing conversion stats file

File is ordered like
Flowcell
  Project
    Sample
      Barcode
        Lane
         Tile
           Raw
             ClusterCount
             Read nr1
             Read nr2
           Pf (passed filter)
             ClusterCount
             Read nr1
             Read nr2


There are many tiles for each lane. We are not interested about the tile level but want to add upp the information to
lane level, just like we do in the demux stats.
"""
import copy
import logging
from pathlib import Path
from typing import Dict, List, Optional, Set
from xml.etree.ElementTree import Element, iterparse

from pydantic.v1 import BaseModel

LOG = logging.getLogger(__name__)


class SampleConversionResults(BaseModel):
    """Class to collect the demultiplexing conversion results for a sample on a lane"""

    raw_cluster_count: int = 0
    raw_yield: int = 0
    pass_filter_cluster_count: int = 0
    pass_filter_read1_yield: int = 0
    pass_filter_read2_yield: int = 0
    pass_filter_yield: int = 0
    pass_filter_q30: int = 0
    pass_filter_read1_q30: int = 0
    pass_filter_read2_q30: int = 0
    pass_filter_quality_score_sum: int = 0
    pass_filter_quality_score: float = 0
    barcode: str = ""
    sample_id: str = ""


class UnknownBarcode(BaseModel):
    barcode: str
    read_count: int


class ConversionStats:
    def __init__(self, conversion_stats_path: Path):
        self.conversion_stats_path: Path = conversion_stats_path
        self._current_sample = ""
        self._current_barcode = ""
        self._results: SampleConversionResults = SampleConversionResults()
        self._skip_entry: bool = False
        self._current_lane: int = 0
        self.unknown_barcodes_entry: bool = False
        # Keep track where we are in the xml file
        self.current_path: List[str] = []
        self.flowcell_id = ""
        self.projects: Set[str] = set()
        self.samples: Set[str] = set()
        self.barcodes: Set[str] = set()
        self.barcode_to_sample: Dict[str, str] = {}
        self.lanes: Set[int] = set()
        self.unknown_barcodes: List[UnknownBarcode] = []
        # Mapping from lane to barcodes and the results
        self.lanes_to_barcode: Dict[int, Dict[str, SampleConversionResults]] = {}
        self.barcode_to_lanes: Dict[str, Dict[int, SampleConversionResults]] = {}
        self.lanes_to_unknown_barcode: Dict[int, List[UnknownBarcode]] = {}
        # This is just a summary of all raw clusters per lane
        self.raw_clusters_per_lane: Dict[int, int] = {}
        self.parse_file()

    @staticmethod
    def get_current_tag(node: Element) -> str:
        current_tag = node.tag
        if current_tag == "Read":
            read_number = "1" if node.attrib["number"] == "1" else "2"
            current_tag = "_".join(["Read", read_number])
        return current_tag

    def parse_file(self) -> None:
        """Parse a file with demux conversion stats information"""
        event: str
        node: Element
        LOG.info("Parsing demux conversion stats file %s", self.conversion_stats_path)
        for event, node in iterparse(str(self.conversion_stats_path), ["start", "end"]):
            # Only search nodes when correct
            current_tag: str = self.get_current_tag(node=node)
            if event == "start":
                self.evaluate_start_event(node=node, current_tag=current_tag)
                continue

            if not self._skip_entry:
                self.evaluate_end_event(node=node, current_tag=current_tag)
            # Remove closed event from current path
            self.current_path.pop()
            # Release element tree from memory
            node.clear()

    @staticmethod
    def update_quality_score(entry: SampleConversionResults) -> None:
        """Calculate the quality score for the lane results"""
        if not entry.pass_filter_yield:
            LOG.debug("Could not find pass filter yield")
            return
        quality_score: float = entry.pass_filter_quality_score_sum / entry.pass_filter_yield
        LOG.debug("Set pass filter quality score to %s", quality_score)
        entry.pass_filter_quality_score = quality_score

    @staticmethod
    def update_summaries(entry: SampleConversionResults) -> None:
        entry.pass_filter_q30 = entry.pass_filter_read1_q30 + entry.pass_filter_read2_q30
        entry.pass_filter_yield = entry.pass_filter_read1_yield + entry.pass_filter_read2_yield

    def update_lane_clusters(self, entry: SampleConversionResults) -> None:
        if self._current_lane not in self.raw_clusters_per_lane:
            self.raw_clusters_per_lane[self._current_lane] = 0
        self.raw_clusters_per_lane[self._current_lane] += entry.raw_cluster_count

    def create_entry(self) -> None:
        entry: SampleConversionResults = copy.deepcopy(self._results)
        self.update_summaries(entry)
        self.update_quality_score(entry)
        self.update_lane_clusters(entry)
        entry.barcode = self._current_barcode
        entry.sample_id = self._current_sample
        if self._current_lane not in self.lanes_to_barcode:
            self.lanes_to_barcode[self._current_lane] = {}
        self.lanes_to_barcode[self._current_lane][self._current_barcode] = entry
        if self._current_barcode not in self.barcode_to_lanes:
            self.barcode_to_lanes[self._current_barcode] = {}
        self.barcode_to_lanes[self._current_barcode][self._current_lane] = entry
        LOG.debug("Adding entry %s for barcode %s", entry, self._current_barcode)
        LOG.debug("Reset conversion results")
        self._results = SampleConversionResults()

    def create_unknown_barcodes_entry(self) -> None:
        LOG.debug("Creating unknown barcode entry for lane %s", self._current_lane)
        lane_results: List[UnknownBarcode] = copy.deepcopy(self.unknown_barcodes)
        self.lanes_to_unknown_barcode[self._current_lane] = lane_results
        self.unknown_barcodes = []
        self.unknown_barcodes_entry = False

    def evaluate_start_event(self, node: Element, current_tag: str) -> None:
        LOG.debug("Add start event %s to current path", current_tag)
        self.current_path.append(current_tag)
        if current_tag == "Lane":
            self.set_current_lane(lane_nr=int(node.attrib["number"]))
        elif current_tag == "Sample":
            sample_name: str = node.attrib["name"]
            if "indexcheck" in sample_name:
                self._skip_entry = True
                LOG.debug("Skip indexcheck sample %s" % sample_name)
                return
            if sample_name.lower() in ["all", "undetermined"]:
                self._skip_entry = True
                LOG.debug("Skip sample %s" % sample_name)
                return
            self._skip_entry = False
            self.set_current_sample(sample_name=sample_name)
        elif current_tag == "Barcode":
            if self.unknown_barcodes_entry:
                unknown_barcode = UnknownBarcode(
                    barcode=node.attrib.get("sequence"), read_count=int(node.attrib.get("count"))
                )
                LOG.debug("Creating unknown barcodes entry for %s", unknown_barcode.barcode)
                self.unknown_barcodes.append(unknown_barcode)
                return
            barcode_sequence: Optional[str] = node.attrib.get("name")
            if not barcode_sequence:
                self._skip_entry = True
                return
            if barcode_sequence.lower() == "all":
                self._skip_entry = True
                return

            self._skip_entry = False
            self.set_current_barcode(barcode=barcode_sequence)
        elif current_tag == "Project":
            self.set_project(project_name=node.attrib["name"])
        elif current_tag == "Flowcell":
            self.set_flowcell_id(node.attrib["flowcell-id"])
        elif current_tag == "TopUnknownBarcodes":
            LOG.debug("Set unknown barcodes entry to true")
            self._skip_entry = False
            self.unknown_barcodes_entry = True

    def evaluate_end_event(self, node: Element, current_tag: str) -> None:
        # We will add the information for each tile
        if "Tile" in self.current_path:
            self.process_tile_info(node=node, current_tag=current_tag)
        elif current_tag == "Lane":
            if self.unknown_barcodes_entry:
                self.create_unknown_barcodes_entry()
            else:
                self.create_entry()

    def update_cluster_count(self, cluster_count: int) -> None:
        if "Raw" in self.current_path:
            self._results.raw_cluster_count += cluster_count
        else:
            self._results.pass_filter_cluster_count += cluster_count

    def update_yield(self, yield_count: int) -> None:
        if "Raw" in self.current_path:
            self._results.raw_yield += yield_count
            return
        if "Read_1" in self.current_path:
            self._results.pass_filter_read1_yield += yield_count
        else:
            self._results.pass_filter_read2_yield += yield_count

    def update_quality_yield(self, yield_count: int) -> None:
        if "Raw" in self.current_path:
            # We don't case about quality yield for raw results
            return
        if "Read_1" in self.current_path:
            self._results.pass_filter_read1_q30 += yield_count
        else:
            self._results.pass_filter_read2_q30 += yield_count

    def update_quality_score_sum(self, quality_score: int) -> None:
        if "Raw" in self.current_path:
            # We don't case about quality score raw results
            return
        self._results.pass_filter_quality_score_sum += quality_score

    def process_tile_info(self, node: Element, current_tag: str) -> None:
        if current_tag == "ClusterCount":
            self.update_cluster_count(cluster_count=int(node.text))
        elif current_tag == "Yield":
            self.update_yield(yield_count=int(node.text))
        elif current_tag == "YieldQ30":
            self.update_quality_yield(yield_count=int(node.text))
        elif current_tag == "QualityScoreSum":
            self.update_quality_score_sum(quality_score=int(node.text))

    def set_current_lane(self, lane_nr: int) -> None:
        LOG.debug("Set current lane to %s", lane_nr)
        self._current_lane = lane_nr
        self.lanes.add(lane_nr)

    def set_current_barcode(self, barcode: str) -> None:
        LOG.debug("Set current barcode %s", barcode)
        self._current_barcode = barcode
        self.barcodes.add(barcode)
        self.barcode_to_sample[barcode] = self._current_sample

    def set_current_sample(self, sample_name: str) -> None:
        LOG.debug("Set current sample to %s", sample_name)
        self._current_sample = sample_name
        self.samples.add(sample_name)

    def set_project(self, project_name: str) -> None:
        LOG.debug("Set current project to %s", project_name)
        self.projects.add(project_name)

    def set_flowcell_id(self, flowcell: str) -> None:
        LOG.debug("Set flowcell id to %s", flowcell)
        self.flowcell_id = flowcell

    def __repr__(self):
        return f"ConversionStats(flowcell_id={self.flowcell_id},projects={self.projects},lanes={self.lanes}"



"""Tests for functions related to dummy samples."""
from cg.apps.demultiplex.sample_sheet.dummy_sample import get_dummy_sample, get_dummy_sample_name
from cg.apps.demultiplex.sample_sheet.index import Index
from cg.apps.demultiplex.sample_sheet.models import (
    FlowCellSample,
    FlowCellSampleNovaSeq6000Bcl2Fastq,
)


def test_get_dummy_sample_name():
    """Test that a dummy sample has the correct name."""
    # GIVEN a raw sample name from the index file
    raw_sample_name = "D10 - D710-D504 (TCCGCGAA-GGCTCTGA)"

    # WHEN converting it to a dummy sample name
    dummy_sample_name: str = get_dummy_sample_name(sample_name=raw_sample_name)

    # THEN the name had spaces and parentheses replaced by dashes
    assert dummy_sample_name == "D10---D710-D504--TCCGCGAA-GGCTCTGA-"


def test_get_dummy_sample(bcl2fastq_flow_cell_id: str, valid_index: Index):
    """Test that a created dummy samples has the correct id."""
    # GIVEN some dummy sample data

    # WHEN creating the dummy sample for a bcl2fastq sample sheet
    dummy_flow_cell_sample: FlowCellSample = get_dummy_sample(
        flow_cell_id=bcl2fastq_flow_cell_id,
        dummy_index=valid_index.sequence,
        lane=1,
        name=valid_index.name,
        sample_type=FlowCellSampleNovaSeq6000Bcl2Fastq,
    )

    # THEN the sample id was correct
    assert dummy_flow_cell_sample.sample_id == get_dummy_sample_name(valid_index.name)



"""CLI support for tower pipelines."""


import click

OPTION_COMPUTE_ENV = click.option(
    "--compute-env",
    type=str,
    help="Compute environment name. If not specified the primary compute environment will be used.",
)
OPTION_TOWER_RUN_ID = click.option(
    "--nf-tower-id",
    type=str,
    is_flag=False,
    default=None,
    help="NF-Tower ID of run to relaunch. If not provided the latest NF-Tower ID for a case will be used.",
)



from pathlib import Path

from cg.apps.cgstats.parsers.conversion_stats import ConversionStats


def test_parse_demux_stats(conversion_stats_path: Path):
    # GIVEN an existing conversion stats file
    assert conversion_stats_path.exists()
    # GIVEN that the file have some content
    demux_info: str = conversion_stats_path.read_text()
    assert demux_info

    # WHEN instantiating a conversion stats parser
    parser: ConversionStats = ConversionStats(conversion_stats_path=conversion_stats_path)

    # THEN assert that the parser have some content
    assert parser.lanes_to_barcode



import collections
from typing import List

from pydantic.v1 import BaseModel, conlist, validator

from cg.constants.nextflow import DELIVER_FILE_HEADERS
from cg.exc import SampleSheetError


class PipelineParameters(BaseModel):
    clusterOptions: str
    priority: str


class NextflowSample(BaseModel):
    """Nextflow samplesheet model.

    Attributes:
        sample: sample name, corresponds to case_id
        fastq_forward: list of all fastq read1 files corresponding to sample
        fastq_reverse: list of all fastq read2 files corresponding to sample
    """

    sample: str
    fastq_forward: conlist(str, min_items=1)
    fastq_reverse: conlist(str, min_items=1)

    @validator("fastq_reverse")
    def fastq_forward_reverse_length_match(
        cls, fastq_reverse: List[str], values: dict
    ) -> List[str]:
        """Verify that the number of fastq forward files is the same as for the reverse."""
        if len(fastq_reverse) != len(values.get("fastq_forward")):
            raise SampleSheetError("Fastq file length for forward and reverse do not match")
        return fastq_reverse


class NextflowDeliverables(BaseModel):
    """Nextflow deliverables model

    Attributes:
        deliverables: dictionary containing format, path, path_index, step, tag and id keys.
    """

    deliverables: dict

    @validator("deliverables")
    def headers(cls, v: dict) -> None:
        """Validate header format."""
        if collections.Counter(list(v.keys())) != collections.Counter(DELIVER_FILE_HEADERS):
            raise ValueError(
                f"Headers are not matching the standard header format: {DELIVER_FILE_HEADERS}"
            )
        for key, value in v.items():
            if not value and key != "path_index":
                raise ValueError("An entry other than path_index is empty!")



from pathlib import Path

import pytest

from cg.constants.demultiplexing import BclConverter
from cg.models.demultiplex.demux_results import DemuxResults
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData


@pytest.fixture(name="demultiplexed_bcl_convert_flow_cell")
def fixture_demultiplexed_bcl_convert_flow_cell(
    demultiplexed_runs: Path, bcl_convert_flow_cell_full_name: str
) -> Path:
    return Path(demultiplexed_runs, bcl_convert_flow_cell_full_name)


@pytest.fixture(name="dragen_demux_results")
def fixture_dragen_demux_results(
    demultiplexed_bcl_convert_flow_cell: Path, bcl_convert_flow_cell: FlowCellDirectoryData
) -> DemuxResults:
    return DemuxResults(
        demux_dir=demultiplexed_bcl_convert_flow_cell,
        flow_cell=bcl_convert_flow_cell,
        bcl_converter=BclConverter.DRAGEN,
    )



"""API for uploading cancer observations."""
import logging
from pathlib import Path
from typing import Dict, List

from housekeeper.store.models import Version, File

from cg.apps.loqus import LoqusdbAPI
from cg.constants.observations import (
    BalsamicObservationsAnalysisTag,
    LoqusdbInstance,
    LOQUSDB_BALSAMIC_SEQUENCING_METHODS,
    BalsamicLoadParameters,
    LOQUSDB_ID,
    LoqusdbBalsamicCustomers,
)
from cg.constants.sequencing import SequencingMethod
from cg.exc import LoqusdbUploadCaseError, CaseNotFoundError, LoqusdbDuplicateRecordError
from cg.meta.observations.observations_api import ObservationsAPI
from cg.models.cg_config import CGConfig
from cg.models.observations.input_files import BalsamicObservationsInputFiles
from cg.store.models import Family
from cg.utils.dict import get_full_path_dictionary

LOG = logging.getLogger(__name__)


class BalsamicObservationsAPI(ObservationsAPI):
    """API to manage Balsamic observations."""

    def __init__(self, config: CGConfig, sequencing_method: SequencingMethod):
        super().__init__(config)
        self.sequencing_method: SequencingMethod = sequencing_method
        self.loqusdb_somatic_api: LoqusdbAPI = self.get_loqusdb_api(LoqusdbInstance.SOMATIC)
        self.loqusdb_tumor_api: LoqusdbAPI = self.get_loqusdb_api(LoqusdbInstance.TUMOR)

    def load_observations(self, case: Family, input_files: BalsamicObservationsInputFiles) -> None:
        """Load observation counts to Loqusdb for a Balsamic case."""
        if self.sequencing_method not in LOQUSDB_BALSAMIC_SEQUENCING_METHODS:
            LOG.error(
                f"Sequencing method {self.sequencing_method} is not supported by Loqusdb. Cancelling upload."
            )
            raise LoqusdbUploadCaseError

        loqusdb_upload_apis: List[LoqusdbAPI] = [self.loqusdb_somatic_api, self.loqusdb_tumor_api]
        for loqusdb_api in loqusdb_upload_apis:
            if self.is_duplicate(
                case=case,
                loqusdb_api=loqusdb_api,
                profile_vcf_path=None,
                profile_threshold=None,
            ):
                LOG.error(f"Case {case.internal_id} has already been uploaded to Loqusdb")
                raise LoqusdbDuplicateRecordError

        for loqusdb_api in loqusdb_upload_apis:
            self.load_cancer_observations(
                case=case, input_files=input_files, loqusdb_api=loqusdb_api
            )

        # Update Statusdb with a germline Loqusdb ID
        loqusdb_id: str = str(self.loqusdb_tumor_api.get_case(case_id=case.internal_id)[LOQUSDB_ID])
        self.update_statusdb_loqusdb_id(samples=case.samples, loqusdb_id=loqusdb_id)

    def load_cancer_observations(
        self, case: Family, input_files: BalsamicObservationsInputFiles, loqusdb_api: LoqusdbAPI
    ) -> None:
        """Load cancer observations to a specific Loqusdb API."""
        is_somatic_db: bool = "somatic" in str(loqusdb_api.config_path)
        is_paired_analysis: bool = len(self.store.get_samples_by_case_id(case.internal_id)) == 2
        if is_somatic_db:
            if not is_paired_analysis:
                return
            LOG.info("Uploading somatic observations to Loqusdb")
            snv_vcf_path: Path = input_files.snv_vcf_path
            sv_vcf_path: Path = input_files.sv_vcf_path
        else:
            LOG.info("Uploading germline observations to Loqusdb")
            snv_vcf_path: Path = input_files.snv_germline_vcf_path
            sv_vcf_path: Path = input_files.sv_germline_vcf_path if is_paired_analysis else None

        load_output: dict = loqusdb_api.load(
            case_id=case.internal_id,
            snv_vcf_path=snv_vcf_path,
            sv_vcf_path=sv_vcf_path,
            qual_gq=True,
            gq_threshold=BalsamicLoadParameters.QUAL_THRESHOLD.value
            if is_somatic_db
            else BalsamicLoadParameters.QUAL_GERMLINE_THRESHOLD.value,
        )
        LOG.info(f"Uploaded {load_output['variants']} variants to {repr(loqusdb_api)}")

    def extract_observations_files_from_hk(
        self, hk_version: Version
    ) -> BalsamicObservationsInputFiles:
        """Extract observations files given a housekeeper version for cancer."""
        input_files: Dict[str, File] = {
            "snv_germline_vcf_path": self.housekeeper_api.files(
                version=hk_version.id, tags=[BalsamicObservationsAnalysisTag.SNV_GERMLINE_VCF]
            ).first(),
            "snv_vcf_path": self.housekeeper_api.files(
                version=hk_version.id, tags=[BalsamicObservationsAnalysisTag.SNV_VCF]
            ).first(),
            "sv_germline_vcf_path": self.housekeeper_api.files(
                version=hk_version.id, tags=[BalsamicObservationsAnalysisTag.SV_GERMLINE_VCF]
            ).first(),
            "sv_vcf_path": self.housekeeper_api.files(
                version=hk_version.id, tags=[BalsamicObservationsAnalysisTag.SV_VCF]
            ).first(),
        }
        return BalsamicObservationsInputFiles(**get_full_path_dictionary(input_files))

    def delete_case(self, case: Family) -> None:
        """Delete cancer case observations from Loqusdb."""
        loqusdb_apis: List[LoqusdbAPI] = [self.loqusdb_somatic_api, self.loqusdb_tumor_api]
        for loqusdb_api in loqusdb_apis:
            if not loqusdb_api.get_case(case.internal_id):
                LOG.error(
                    f"Case {case.internal_id} could not be found in Loqusdb. Skipping case deletion."
                )
                raise CaseNotFoundError

        for loqusdb_api in loqusdb_apis:
            loqusdb_api.delete_case(case.internal_id)
        self.update_statusdb_loqusdb_id(samples=case.samples, loqusdb_id=None)
        LOG.info(f"Removed observations for case {case.internal_id} from Loqusdb")

    def get_loqusdb_customers(self) -> LoqusdbBalsamicCustomers:
        """Returns the customers that are entitled to Cancer Loqusdb uploads."""
        return LoqusdbBalsamicCustomers



import click

from cg.constants.constants import GenomeVersion
from cg.constants.priority import SlurmQos
from cg.constants.subject import Gender

OPTION_GENDER = click.option(
    "--gender",
    type=click.Choice([Gender.FEMALE, Gender.MALE]),
    required=False,
    help="Case associated gender. Set this option to override the one selected by the customer in StatusDB.",
)
OPTION_GENOME_VERSION = click.option(
    "--genome-version",
    show_default=True,
    default=GenomeVersion.hg19,
    type=click.Choice([GenomeVersion.hg19, GenomeVersion.hg38, GenomeVersion.canfam3]),
    help="Type and build version of the reference genome. Set this option to override the default.",
)
OPTION_PANEL_BED = click.option(
    "--panel-bed",
    required=False,
    help="Panel BED is determined based on capture kit \
    used for library prep. Set this option to override the default",
)
OPTION_RUN_ANALYSIS = click.option(
    "-r",
    "--run-analysis",
    is_flag=True,
    default=False,
    help="Execute BALSAMIC in non-dry mode",
)
OPTION_QOS = click.option(
    "-qos",
    "--slurm-quality-of-service",
    type=click.Choice([SlurmQos.LOW, SlurmQos.NORMAL, SlurmQos.HIGH, SlurmQos.EXPRESS]),
    help="Job priority in SLURM. Setting this option will override the StatusDB case priority.",
)
OPTION_PON_CNN = click.option(
    "--pon-cnn",
    type=click.Path(exists=True),
    required=False,
    help="Panel of normal reference (.cnn) for CNVkit",
)
OPTION_OBSERVATIONS = click.option(
    "--observations",
    type=click.Path(exists=True),
    multiple=True,
    required=False,
    help="VCF paths of clinical and/or cancer SNVs and SVs observations (WGS analysis only). Set this option to "
    "override the latest Loqusdb dump files.",
)



"""Module for reading and writing xml files."""

import xml.etree.ElementTree as ET

from cg.constants import FileExtensions
from cg.io.validate_path import validate_file_suffix
from pathlib import Path


def read_xml(file_path: Path) -> ET.ElementTree:
    """Read content in a xml file to an ElementTree."""
    validate_file_suffix(path_to_validate=file_path, target_suffix=FileExtensions.XML)
    tree = ET.parse(file_path)
    return tree


def write_xml(tree: ET.ElementTree, file_path: Path) -> None:
    """Write content to a xml file."""
    tree.write(file_path, encoding="utf-8", xml_declaration=True)



import re
from pathlib import Path

from cg.constants.constants import FileExtensions
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.exc import FlowCellError


def is_valid_sample_fastq_file(sample_fastq: Path, sample_internal_id: str) -> bool:
    """
    Validate that the sample fastq file name is formatted as expected.

    Assumptions:
    1. The sample fastq file name ends with .fastq.gz
    2. The sample fastq file name contains the lane number formatted as _L<lane_number>
    3. The sample internal id is present in the parent directory name or in the file name.
    """
    sample_id_in_directory: bool = is_sample_id_in_directory_name(
        directory=sample_fastq.parent, sample_internal_id=sample_internal_id
    )
    sample_id_in_file_name: bool = is_sample_id_in_file_name(
        sample_fastq=sample_fastq, sample_internal_id=sample_internal_id
    )

    return (
        is_file_path_compressed_fastq(sample_fastq)
        and is_lane_in_fastq_file_name(sample_fastq)
        and (sample_id_in_directory or sample_id_in_file_name)
    )


def is_file_path_compressed_fastq(file_path: Path) -> bool:
    return file_path.name.endswith(f"{FileExtensions.FASTQ}{FileExtensions.GZIP}")


def is_lane_in_fastq_file_name(sample_fastq: Path) -> bool:
    """Validate that fastq contains lane number formatted as _L<lane_number>"""
    return bool(re.search(r"_L\d+", sample_fastq.name))


def is_sample_id_in_directory_name(directory: Path, sample_internal_id: str) -> bool:
    """Validate that directory name contains the sample id formatted as Sample_<sample_id> or Sample_<sample_id>_."""
    sample_pattern: str = f"Sample_{sample_internal_id}"
    return f"{sample_pattern}_" in directory.name or sample_pattern == directory.name


def is_sample_id_in_file_name(sample_fastq: Path, sample_internal_id: str) -> bool:
    """Validate that file name contains the sample id formatted as <sample_id>_."""
    return f"{sample_internal_id}_" in sample_fastq.name


def is_bcl2fastq_demux_folder_structure(flow_cell_directory: Path) -> bool:
    """Check if flow cell directory is a Bcl2fastq demux folder structure."""
    for folder in flow_cell_directory.glob(pattern="*"):
        if re.search(DemultiplexingDirsAndFiles.BCL2FASTQ_TILE_DIR_PATTERN.value, str(folder)):
            return True
    return False


def is_demultiplexing_complete(flow_cell_directory: Path) -> bool:
    return Path(flow_cell_directory, DemultiplexingDirsAndFiles.DEMUX_COMPLETE).exists()


def is_flow_cell_ready_for_delivery(flow_cell_directory: Path) -> bool:
    return Path(flow_cell_directory, DemultiplexingDirsAndFiles.DELIVERY).exists()


def validate_sample_sheet_exists(flow_cell_run_directory: Path) -> None:
    sample_sheet_path: Path = Path(
        flow_cell_run_directory, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME
    )
    if not sample_sheet_path.exists():
        raise FlowCellError(
            f"Sample sheet {sample_sheet_path} does not exist in flow cell run directory."
        )


def validate_demultiplexing_complete(flow_cell_output_directory: Path) -> None:
    if not is_demultiplexing_complete(flow_cell_output_directory):
        raise FlowCellError(
            f"Demultiplexing not completed for flow cell directory {flow_cell_output_directory}."
        )


def validate_flow_cell_delivery_status(flow_cell_output_directory: Path, force: bool) -> None:
    if is_flow_cell_ready_for_delivery(flow_cell_output_directory) and not force:
        raise FlowCellError(
            f"Flow cell output directory {flow_cell_output_directory} has already been processed and is ready for delivery."
        )
    return


def is_flow_cell_ready_for_postprocessing(
    flow_cell_output_directory: Path, flow_cell_run_directory: Path, force: bool = False
) -> None:
    validate_sample_sheet_exists(flow_cell_run_directory)
    validate_demultiplexing_complete(flow_cell_output_directory)
    validate_flow_cell_delivery_status(flow_cell_output_directory, force=force)



import logging
from pydantic.v1 import BaseModel, Extra, Field
from typing import List

from cg.constants.constants import GenomeVersion
from cg.constants.demultiplexing import SampleSheetNovaSeq6000Sections, SampleSheetNovaSeqXSections

LOG = logging.getLogger(__name__)


class FlowCellSample(BaseModel):
    """Base class for flow cell samples."""

    lane: int
    sample_id: str
    index: str
    index2: str = ""

    class Config:
        allow_population_by_field_name = True
        extra = Extra.ignore


class FlowCellSampleNovaSeq6000(FlowCellSample):
    """Base class for NovaSeq6000 flow cell samples."""

    flowcell_id: str = Field("", alias=SampleSheetNovaSeq6000Sections.Data.FLOW_CELL_ID.value)
    lane: int = Field(..., alias=SampleSheetNovaSeq6000Sections.Data.LANE.value)
    sample_ref: str = Field(
        GenomeVersion.hg19.value, alias=SampleSheetNovaSeq6000Sections.Data.SAMPLE_REFERENCE.value
    )
    index: str = Field(..., alias=SampleSheetNovaSeq6000Sections.Data.INDEX_1.value)
    index2: str = Field("", alias=SampleSheetNovaSeq6000Sections.Data.INDEX_2.value)
    sample_name: str = Field(..., alias=SampleSheetNovaSeq6000Sections.Data.SAMPLE_NAME.value)
    control: str = Field("N", alias=SampleSheetNovaSeq6000Sections.Data.CONTROL.value)
    recipe: str = Field("R1", alias=SampleSheetNovaSeq6000Sections.Data.RECIPE.value)
    operator: str = Field("script", alias=SampleSheetNovaSeq6000Sections.Data.OPERATOR.value)


class FlowCellSampleNovaSeqX(FlowCellSample):
    """Class that represents a NovaSeqX flow cell sample."""

    lane: int = Field(..., alias=SampleSheetNovaSeqXSections.Data.LANE.value)
    sample_id: str = Field(..., alias=SampleSheetNovaSeqXSections.Data.SAMPLE_INTERNAL_ID.value)
    index: str = Field(..., alias=SampleSheetNovaSeqXSections.Data.INDEX_1.value)
    index2: str = Field("", alias=SampleSheetNovaSeqXSections.Data.INDEX_2.value)
    adapter_read_1: str = Field("", alias=SampleSheetNovaSeqXSections.Data.ADAPTER_READ_1.value)
    adapter_read_2: str = Field("", alias=SampleSheetNovaSeqXSections.Data.ADAPTER_READ_2.value)
    barcode_mismatches_1: int = Field(
        1, alias=SampleSheetNovaSeqXSections.Data.BARCODE_MISMATCHES_1.value
    )
    barcode_mismatches_2: int = Field(
        1, alias=SampleSheetNovaSeqXSections.Data.BARCODE_MISMATCHES_2.value
    )


class FlowCellSampleNovaSeq6000Bcl2Fastq(FlowCellSampleNovaSeq6000):
    """Class that represents a NovaSeq6000 Bcl2fastq flow cell sample."""

    sample_id: str = Field(
        ..., alias=SampleSheetNovaSeq6000Sections.Data.SAMPLE_INTERNAL_ID_BCL2FASTQ.value
    )
    project: str = Field(
        ..., alias=SampleSheetNovaSeq6000Sections.Data.SAMPLE_PROJECT_BCL2FASTQ.value
    )


class FlowCellSampleNovaSeq6000Dragen(FlowCellSampleNovaSeq6000):
    """Class that represents a NovaSeq6000 Dragen flow cell sample."""

    sample_id: str = Field(
        ..., alias=SampleSheetNovaSeq6000Sections.Data.SAMPLE_INTERNAL_ID_BCLCONVERT.value
    )
    project: str = Field(
        ..., alias=SampleSheetNovaSeq6000Sections.Data.SAMPLE_PROJECT_BCLCONVERT.value
    )


class SampleSheet(BaseModel):
    samples: List[FlowCellSample]


class SampleSheetBcl2Fastq(SampleSheet):
    samples: List[FlowCellSampleNovaSeq6000Bcl2Fastq]


class SampleSheetDragen(SampleSheet):
    samples: List[FlowCellSampleNovaSeq6000Dragen]



from pathlib import Path
from typing import List

import pytest

from cg.apps.demultiplex.sample_sheet.index import Index
from cg.apps.demultiplex.sample_sheet.sample_sheet_creator import (
    SampleSheetCreatorV1,
    SampleSheetCreatorV2,
)
from cg.apps.demultiplex.sample_sheet.models import (
    FlowCellSampleNovaSeq6000Bcl2Fastq,
    FlowCellSampleNovaSeq6000Dragen,
    FlowCellSampleNovaSeqX,
)
from cg.constants.demultiplexing import BclConverter, SampleSheetNovaSeq6000Sections
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData


@pytest.fixture(name="output_dirs_bcl2fastq")
def fixture_output_dirs_bcl2fastq(demultiplexed_runs: Path) -> Path:
    """Return the output path a dir with flow cells that have finished demultiplexing using
    bcl2fastq."""
    return Path(demultiplexed_runs, BclConverter.BCL2FASTQ)


@pytest.fixture(name="demux_run_dir_bcl2fastq")
def fixture_demux_run_dir_bcl2fastq(flow_cell_runs_dir: Path) -> Path:
    """Return the path to a dir with flowcells ready for demultiplexing"""
    return Path(flow_cell_runs_dir, BclConverter.BCL2FASTQ)


@pytest.fixture(name="demux_run_dir_dragen")
def fixture_demux_run_dir_dragen(flow_cell_runs_dir: Path) -> Path:
    """Return the path to a dir with flowcells ready for demultiplexing"""
    return Path(flow_cell_runs_dir, BclConverter.DRAGEN)


@pytest.fixture(name="index_obj")
def fixture_index_obj() -> Index:
    return Index(name="C07 - UDI0051", sequence="AACAGGTT-ATACCAAG")


@pytest.fixture(name="valid_index")
def fixture_valid_index_() -> Index:
    """Return a valid index."""
    return Index(name="C07 - UDI0051", sequence="AACAGGTT-ATACCAAG")


@pytest.fixture(name="lims_novaseq_bcl2fastq_samples")
def fixture_lims_novaseq_bcl2fastq_samples(
    lims_novaseq_samples_raw: List[dict],
) -> List[FlowCellSampleNovaSeq6000Bcl2Fastq]:
    """Return a list of parsed Bcl2fastq flow cell samples"""
    return [FlowCellSampleNovaSeq6000Bcl2Fastq(**sample) for sample in lims_novaseq_samples_raw]


@pytest.fixture(name="lims_novaseq_dragen_samples")
def fixture_lims_novaseq_dragen_samples(
    lims_novaseq_samples_raw: List[dict],
) -> List[FlowCellSampleNovaSeq6000Dragen]:
    """Return a list of parsed Dragen flow cell samples"""
    return [FlowCellSampleNovaSeq6000Dragen(**sample) for sample in lims_novaseq_samples_raw]


@pytest.fixture(name="lims_novaseq_x_samples")
def fixture_lims_novaseq_x_samples(
    lims_novaseq_samples_raw: List[dict],
) -> List[FlowCellSampleNovaSeqX]:
    """Return a list of parsed NovaSeqX flow cell samples"""
    return [FlowCellSampleNovaSeqX(**sample) for sample in lims_novaseq_samples_raw]


@pytest.fixture(name="novaseq_bcl2fastq_sample_sheet_creator")
def fixture_novaseq_bcl2fastq_sample_sheet_creator(
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    lims_novaseq_bcl2fastq_samples: List[FlowCellSampleNovaSeq6000Bcl2Fastq],
) -> SampleSheetCreatorV1:
    """Returns a sample sheet creator for version 1 sample sheets with bcl2fastq format."""
    return SampleSheetCreatorV1(
        flow_cell=bcl2fastq_flow_cell,
        lims_samples=lims_novaseq_bcl2fastq_samples,
        bcl_converter=BclConverter.BCL2FASTQ,
    )


@pytest.fixture(name="novaseq_dragen_sample_sheet_creator")
def fixture_novaseq_dragen_sample_sheet_creator(
    bcl_convert_flow_cell: FlowCellDirectoryData,
    lims_novaseq_dragen_samples: List[FlowCellSampleNovaSeq6000Dragen],
) -> SampleSheetCreatorV1:
    """Returns a sample sheet creator for version 1 sample sheets with dragen format."""
    return SampleSheetCreatorV1(
        flow_cell=bcl_convert_flow_cell,
        lims_samples=lims_novaseq_dragen_samples,
        bcl_converter=BclConverter.DRAGEN,
    )


@pytest.fixture(name="novaseq_x_sample_sheet_creator")
def fixture_novaseq_x_sample_sheet_creator(
    novaseq_x_flow_cell: FlowCellDirectoryData,
    lims_novaseq_x_samples: List[FlowCellSampleNovaSeqX],
) -> SampleSheetCreatorV2:
    """Returns a sample sheet creator for version 2 sample sheets."""
    return SampleSheetCreatorV2(
        flow_cell=novaseq_x_flow_cell,
        lims_samples=lims_novaseq_x_samples,
        bcl_converter=BclConverter.DRAGEN,
    )


# Sample sheet validation


@pytest.fixture(name="sample_sheet_line_sample_1")
def fixture_sample_sheet_line_sample_1() -> List[str]:
    """Return the line in the sample sheet corresponding to a sample."""
    return [
        "HWHMWDMXX",
        "1",
        "ACC7628A68",
        "hg19",
        "ATTCCACACT",
        "TGGTCTTGTT",
        "814206",
        "N",
        "R1",
        "script",
        "814206",
    ]


@pytest.fixture(name="sample_sheet_line_sample_2")
def fixture_sample_sheet_line_sample_2() -> List[str]:
    """Return the line in the sample sheet corresponding to a sample."""
    return [
        "HWHMWDMXX",
        "1",
        "ACC7628A1",
        "hg19",
        "AGTTAGCTGG",
        "GATGAGAATG",
        "814206",
        "N",
        "R1",
        "script",
        "814206",
    ]


@pytest.fixture(name="sample_sheet_bcl2fastq_data_header")
def fixture_sample_sheet_bcl2fastq_data_header() -> List[List[str]]:
    """Return the content of a Bcl2fastq sample sheet data header without samples."""
    return [
        [SampleSheetNovaSeq6000Sections.Data.HEADER],
        [
            SampleSheetNovaSeq6000Sections.Data.FLOW_CELL_ID.value,
            SampleSheetNovaSeq6000Sections.Data.LANE.value,
            SampleSheetNovaSeq6000Sections.Data.SAMPLE_INTERNAL_ID_BCL2FASTQ.value,
            SampleSheetNovaSeq6000Sections.Data.SAMPLE_REFERENCE.value,
            SampleSheetNovaSeq6000Sections.Data.INDEX_1.value,
            SampleSheetNovaSeq6000Sections.Data.INDEX_2.value,
            SampleSheetNovaSeq6000Sections.Data.SAMPLE_NAME.value,
            SampleSheetNovaSeq6000Sections.Data.CONTROL.value,
            SampleSheetNovaSeq6000Sections.Data.RECIPE.value,
            SampleSheetNovaSeq6000Sections.Data.OPERATOR.value,
            SampleSheetNovaSeq6000Sections.Data.SAMPLE_PROJECT_BCL2FASTQ.value,
        ],
    ]


@pytest.fixture(name="sample_sheet_samples_no_header")
def fixture_sample_sheet_no_sample_header(
    sample_sheet_line_sample_1: List[str], sample_sheet_line_sample_2: List[str]
) -> List[List[str]]:
    """Return the content of a sample sheet with samples but without a sample header."""
    return [
        [SampleSheetNovaSeq6000Sections.Data.HEADER],
        sample_sheet_line_sample_1,
        sample_sheet_line_sample_2,
    ]


@pytest.fixture(name="valid_sample_sheet_bcl2fastq")
def fixture_valid_sample_sheet_bcl2fastq(
    sample_sheet_bcl2fastq_data_header: List[List[str]],
    sample_sheet_line_sample_1: List[str],
    sample_sheet_line_sample_2: List[str],
) -> List[List[str]]:
    """Return the content of a valid Bcl2fastq sample sheet."""
    return sample_sheet_bcl2fastq_data_header + [
        sample_sheet_line_sample_1,
        sample_sheet_line_sample_2,
    ]


@pytest.fixture(name="sample_sheet_bcl2fastq_duplicate_same_lane")
def fixture_sample_sheet_bcl2fastq_duplicate_same_lane(
    valid_sample_sheet_bcl2fastq: List[List[str]], sample_sheet_line_sample_2: List[str]
) -> List[List[str]]:
    """Return the content of a Bcl2fastq sample sheet with a duplicated sample in the same lane."""
    valid_sample_sheet_bcl2fastq.append(sample_sheet_line_sample_2)
    return valid_sample_sheet_bcl2fastq


@pytest.fixture(name="sample_sheet_bcl2fastq_duplicate_different_lane")
def fixture_sample_sheet_bcl2fastq_duplicate_different_lane(
    valid_sample_sheet_bcl2fastq: List[List[str]],
) -> List[List[str]]:
    """Return the content of a Bcl2fastq sample sheet with a duplicated sample in a different lane."""
    valid_sample_sheet_bcl2fastq.append(
        [
            "HWHMWDMXX",
            "2",
            "ACC7628A1",
            "hg19",
            "AGTTAGCTGG",
            "GATGAGAATG",
            "814206",
            "N",
            "R1",
            "script",
            "814206",
        ]
    )
    return valid_sample_sheet_bcl2fastq


@pytest.fixture(name="valid_sample_sheet_dragen")
def fixture_valid_sample_sheet_dragen(
    sample_sheet_line_sample_1: List[str], sample_sheet_line_sample_2: List[str]
) -> List[List[str]]:
    """Return the content of a valid Dragen sample sheet."""
    return [
        [SampleSheetNovaSeq6000Sections.Data.HEADER],
        [
            SampleSheetNovaSeq6000Sections.Data.FLOW_CELL_ID.value,
            "Lane",
            "Sample_ID",
            "SampleRef",
            "index",
            "index2",
            "SampleName",
            "Control",
            "Recipe",
            "Operator",
            "Sample_Project",
        ],
        sample_sheet_line_sample_1,
        sample_sheet_line_sample_2,
    ]


@pytest.fixture(name="sample_sheet_dragen_duplicate_same_lane")
def fixture_sample_sheet_dragen_duplicate_same_lane(
    valid_sample_sheet_dragen: List[List[str]], sample_sheet_line_sample_2: List[str]
) -> List[List[str]]:
    """Return the content of a Dragen sample sheet with a duplicated sample in the same lane."""
    valid_sample_sheet_dragen.append(sample_sheet_line_sample_2)
    return valid_sample_sheet_dragen


@pytest.fixture(name="sample_sheet_dragen_duplicate_different_lane")
def fixture_sample_sheet_dragen_duplicate_different_lane(
    valid_sample_sheet_dragen: List[List[str]],
) -> List[List[str]]:
    """Return the content of a Dragen sample sheet with a duplicated sample in a different lane."""
    valid_sample_sheet_dragen.append(
        [
            "HWHMWDMXX",
            "2",
            "ACC7628A1",
            "hg19",
            "AGTTAGCTGG",
            "GATGAGAATG",
            "814206",
            "N",
            "R1",
            "script",
            "814206",
        ]
    )
    return valid_sample_sheet_dragen


@pytest.fixture(name="valid_sample_sheet_bcl2fastq_path")
def fixture_valid_sample_sheet_bcl2fastq_path() -> Path:
    """Return the path to a NovaSeq S2 sample sheet, used in bcl2fastq demultiplexing."""
    return Path("tests", "fixtures", "apps", "demultiplexing", "SampleSheetS2_Bcl2Fastq.csv")


@pytest.fixture(name="valid_sample_sheet_dragen_path")
def fixture_valid_sample_sheet_dragen_path() -> Path:
    """Return the path to a NovaSeq S2 sample sheet, used in dragen demultiplexing."""
    return Path("tests", "fixtures", "apps", "demultiplexing", "SampleSheetS2_Dragen.csv")


@pytest.fixture(name="novaseq6000_flow_cell_sample_1")
def fixture_novaseq6000_flow_cell_sample_1() -> FlowCellSampleNovaSeq6000Bcl2Fastq:
    """Return a NovaSeq sample."""
    return FlowCellSampleNovaSeq6000Bcl2Fastq(
        FCID="HWHMWDMXX",
        Lane=1,
        SampleID="ACC7628A68",
        SampleRef="hg19",
        index="ATTCCACACT",
        index2="TGGTCTTGTT",
        SampleName="814206",
        Control="N",
        Recipe="R1",
        Operator="script",
        Project="814206",
    )


@pytest.fixture(name="novaseq6000_flow_cell_sample_2")
def fixture_novaseq6000_flow_cell_sample_2() -> FlowCellSampleNovaSeq6000Bcl2Fastq:
    """Return a NovaSeq sample."""
    return FlowCellSampleNovaSeq6000Bcl2Fastq(
        FCID="HWHMWDMXX",
        Lane=2,
        SampleID="ACC7628A1",
        SampleRef="hg19",
        index="ATTCCACACT",
        index2="TGGTCTTGTT",
        SampleName="814206",
        Control="N",
        Recipe="R1",
        Operator="script",
        Project="814206",
    )


@pytest.fixture(name="novaseq_x_flow_cell_sample_before_adapt_indexes")
def fixture_novaseq_x_flow_cell_sample_before_adapt_indexes() -> FlowCellSampleNovaSeqX:
    """Return a NovaSeqX sample."""
    return FlowCellSampleNovaSeqX(
        Lane=2,
        Sample_ID="ACC7628A1",
        index="ATTCCACACT-TGGTCTTGTT",
    )


@pytest.fixture(name="novaseq6000_flow_cell_sample_no_dual_index")
def fixture_novaseq6000_flow_cell_sample_no_dual_index() -> FlowCellSampleNovaSeq6000Bcl2Fastq:
    """Return a NovaSeq sample without dual indexes."""
    return FlowCellSampleNovaSeq6000Bcl2Fastq(
        FCID="HWHMWDMXX",
        Lane=2,
        SampleID="ACC7628A1",
        index="ATTCCACACT",
        SampleName="814206",
        Project="814206",
    )


@pytest.fixture(name="novaseq6000_flow_cell_sample_before_adapt_indexes")
def fixture_novaseq6000_flow_cell_sample_before_adapt_indexes() -> (
    FlowCellSampleNovaSeq6000Bcl2Fastq
):
    """Return a NovaSeq sample without dual indexes."""
    return FlowCellSampleNovaSeq6000Bcl2Fastq(
        FCID="HWHMWDMXX",
        Lane=2,
        SampleID="ACC7628A1",
        index="ATTCCACACT-TGGTCTTGTT",
        SampleName="814206",
        Project="814206",
    )



from cg.store import Store
from cg.store.models import Collaboration
from cg.store.filters.status_collaboration_filters import filter_collaboration_by_internal_id


def test_get_collaboration_by_internal_id(base_store: Store, collaboration_id: str):
    """Test getting collaboration by internal id."""
    # GIVEN a store with a collaboration

    # WHEN retrieving the collaboration
    collaboration: Collaboration = filter_collaboration_by_internal_id(
        collaborations=base_store._get_query(table=Collaboration),
        internal_id=collaboration_id,
    ).first()

    # THEN collaboration should be returned
    assert isinstance(collaboration, Collaboration)

    # THEN the internal id should match the fixture
    assert collaboration.internal_id == collaboration_id


def test_get_collaboration_by_internal_id_wrong_name(base_store: Store, collaboration_id: str):
    """Test getting a collaboration with non-existing internal_id."""
    # GIVEN a store a collaboration

    # WHEN attempting to retrieve a non-existing collaboration
    collaboration: Collaboration = filter_collaboration_by_internal_id(
        collaborations=base_store._get_query(table=Collaboration),
        internal_id="missing_collaboration",
    ).first()

    # THEN no collaboration should be returned
    assert collaboration is None
